import pathlib
import shutil
from unittest import TestCase

import os
import subprocess
import unittest

from wopmars.utils.OptionManager import OptionManager
from wopmars import WopMars
from wopmars.Base import Base
from wopmars.SQLManager import SQLManager
from wopmars.models.Execution import Execution
from wopmars.models.TableModificationTime import TableModificationTime
from wopmars.utils.PathManager import PathManager
from wopmars.utils.various import get_current_time


class TestWopmars(TestCase):

    def setUp(self):
        self.test_path = PathManager.get_test_path()  # Get tests path
        OptionManager.initial_test_setup()  # Set tests arguments
        self.__db_url = OptionManager.instance()["--database"]

        self.__example_def_file1 = os.path.join(self.test_path, "resource/wopfile/example_def_file1.yml")
        self.__example_def_file1_only_database = \
            os.path.join(self.test_path, "resource/wopfile/example_def_file1_only_database.yml")
        self.__example_def_file2_only_files = os.path.join(self.test_path, "resource/wopfile/example_def_file2_only_files.yml")
        self.__example_def_file4 = os.path.join(self.test_path, "resource/wopfile/example_def_file4.yml")
        self.__example_def_file5_never_ready = os.path.join(self.test_path, "resource/wopfile/example_def_file5_never_ready.yml")
        self.__example_def_file_input_not_ready = os.path.join(self.test_path, "resource/wopfile/example_def_file_input_not_ready.yml")

    def tearDown(self):
        SQLManager.instance().get_session().close()
        SQLManager.instance().drop_all()
        shutil.rmtree(os.path.join(self.test_path, "outdir"), ignore_errors=True)
        OptionManager._drop()
        SQLManager._drop()

    def test_run(self):
        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file1, "-v", "-d", self.test_path]
        with self.assertRaises(SystemExit) as se:
            WopMars().run(cmd_line)
        self.assertTrue(os.path.exists(os.path.join(self.test_path, 'outdir/output_file1.txt')))
        self.assertTrue(os.path.exists(os.path.join(self.test_path, 'outdir/output_file2.txt')))
        self.assertTrue(os.path.exists(os.path.join(self.test_path, 'outdir/output_file7.txt')))
        # self.assertEqual(se.exception.code, 0)

    def test_run_touch(self):

        """This is a tests for the touch function"""

        ################################################################################################################
        #
        # Normal run until rule 2
        #
        ################################################################################################################

        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file1, "-v", "-d", self.test_path, "-U", "rule2"]
        with self.assertRaises(SystemExit) as se:
            WopMars().run(cmd_line)
        # self.assertTrue(os.path.exists(os.path.join(self.test_path, 'outdir/output_file1.txt')))
        # self.assertEqual(se.exception.code, 0)
        output_file1_mtime = os.path.getmtime(os.path.join(self.test_path, 'outdir/output_file1.txt'))
        foobase_mtime = SQLManager.instance().get_session().query(TableModificationTime)\
            .filter_by(table_name='FooBase').first().mtime_epoch_millis

        ################################################################################################################
        #
        # Dry-run
        #
        ################################################################################################################

        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file1, "-v",
                    "-d", self.test_path, "-U", "rule2 --dry-run"]
        with self.assertRaises(SystemExit) as se:
            WopMars().run(cmd_line)
        # self.assertTrue(os.path.exists(os.path.join(self.test_path, 'outdir/output_file1.txt')))
        # self.assertEqual(se.exception.code, 0)
        output_file1_mtime_after_dry = os.path.getmtime(os.path.join(self.test_path, 'outdir/output_file1.txt'))
        foobase_mtime_after_dry = SQLManager.instance().get_session().query(TableModificationTime)\
            .filter_by(table_name='FooBase').first().mtime_epoch_millis

        ################################################################################################################
        #
        # Touch output files
        #
        ################################################################################################################

        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file1, "-v", "-d", self.test_path,
                    "-U", "rule2", "--touch"]
        with self.assertRaises(SystemExit) as se:
            WopMars().run(cmd_line)
        output_file1_mtime_after_touch = os.path.getmtime(os.path.join(self.test_path, 'outdir/output_file1.txt'))
        foobase_mtime_after_touch = SQLManager.instance().get_session().query(TableModificationTime)\
            .filter_by(table_name='FooBase').first().mtime_epoch_millis

        ################################################################################################################
        #
        # Assert that modification times have been updated
        #
        ################################################################################################################

        self.assertEqual(output_file1_mtime, output_file1_mtime_after_dry)
        self.assertEqual(foobase_mtime, foobase_mtime_after_dry)

        self.assertLess(output_file1_mtime, output_file1_mtime_after_touch)
        self.assertLess(foobase_mtime, foobase_mtime_after_touch)

    def test_db_forceall(self):
        """Tests the forceall option in the database. Will carry out different runs in either default or forceall mode
        that will either increase or keep same the number of rows of the FooBase table"""

        pathlib.Path(os.path.join(self.test_path, "outdir/output_file1.txt")).touch()

        ################################################################################################################
        #
        # Run 1: normal mode. Results in 1000 rows
        #
        ################################################################################################################

        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file1_only_database, "-v",
                    "-d", self.test_path]
        with self.assertRaises(SystemExit) as se:
            WopMars().run(cmd_line)
        foobase_model = Base.metadata.tables['FooBase']
        row_count = SQLManager.instance().get_session().query(foobase_model).count()
        self.assertTrue(row_count == 1000)
        self.assertEqual(se.exception.code, 0)

        ################################################################################################################
        #
        # Run 2: normal mode. Still 1000 rows, because it does not execute
        #
        ################################################################################################################

        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file1_only_database, "-v",
                    "-d", self.test_path]
        with self.assertRaises(SystemExit) as se:
            WopMars().run(cmd_line)
        row_count = SQLManager.instance().get_session().query(foobase_model).count()
        self.assertTrue(row_count == 1000)
        self.assertEqual(se.exception.code, 0)

        ################################################################################################################
        #
        # Run 3: forceall mode. Now 2000 rows, because it is forced to execute
        #
        ################################################################################################################

        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file1_only_database, "-v",
                    "-d", self.test_path, "--forceall"]
        with self.assertRaises(SystemExit) as se:
            WopMars().run(cmd_line)
        row_count = SQLManager.instance().get_session().query(foobase_model).count()
        self.assertTrue(row_count == 2000)
        self.assertEqual(se.exception.code, 0)

        ################################################################################################################
        #
        # Run 4: normal mode. Again 2000 rows, because it does not execute
        #
        ################################################################################################################

        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file1_only_database, "-v",
                    "-d", self.test_path]
        with self.assertRaises(SystemExit) as se:
            WopMars().run(cmd_line)
        row_count = SQLManager.instance().get_session().query(foobase_model).count()
        self.assertTrue(row_count == 2000)
        self.assertEqual(se.exception.code, 0)

    def test_forceall_dryrun(self):
        """This tests carries out a normal a forceall and a dryrun and stores the modification times.
        Then it will tests that the modification time of the two first outputs are different and the
         second and third equal"""
        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file1, "-v", "-d", self.test_path]
        with self.assertRaises(SystemExit) as se:
            WopMars().run(cmd_line)
        # Test that modification date of first run is different from second forced run
        mtime1 = os.path.getmtime(os.path.join(self.test_path, 'outdir/output_file7.txt'))
        # Run forceall
        cmd_line_force = ["python", "-D", self.__db_url, "-w", self.__example_def_file1, "-v",
                    "-d", self.test_path, "--forceall"]
        with self.assertRaises(SystemExit) as se:
            WopMars().run(cmd_line_force)
        mtime2 = os.path.getmtime(os.path.join(self.test_path, 'outdir/output_file7.txt'))
        # Run dry-run
        cmd_line_dry_run = ["python", "-D", self.__db_url, "-w", self.__example_def_file1, "-v",
                    "-d", self.test_path, "--dry-run"]
        with self.assertRaises(SystemExit) as se:
            WopMars().run(cmd_line_dry_run)
        mtime3 = os.path.getmtime(os.path.join(self.test_path, 'outdir/output_file7.txt'))
        # Assert first two modification times are different
        # Assert second and third modification times same
        self.assertTrue(mtime1 != mtime2)
        self.assertTrue(mtime2 == mtime3)
        self.assertEqual(se.exception.code, 0)

    def test_dry_run(self):
        cmd_line = ["python", "--dry-run", "-D", self.__db_url, "-w", self.__example_def_file1, "-v", "-d",
                    self.test_path]
        with self.assertRaises(SystemExit) as se:
            WopMars().run(cmd_line)
        # The tests is that these files do not exist
        self.assertFalse(os.path.exists(os.path.join(self.test_path, 'outdir/output_file1.txt')))
        self.assertFalse(os.path.exists(os.path.join(self.test_path, 'outdir/output_file2.txt')))
        self.assertFalse(os.path.exists(os.path.join(self.test_path, 'outdir/output_file7.txt')))
        self.assertEqual(se.exception.code, 0)

    def test_run_that_fails(self):
        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file5_never_ready, "-vv", "-d",
                    self.test_path]
        with self.assertRaises(SystemExit) as se:
            WopMars().run(cmd_line)
        self.assertFalse(os.path.exists(os.path.join(self.test_path, 'resources/never_done.txt')))
        self.assertEqual(se.exception.code, 1)

    def test_run_sourcerule_fail(self):
        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file1, "-v", "--since", "failure"]
        with self.assertRaises(SystemExit) as se:
            WopMars().run(cmd_line)
        self.assertFalse(os.path.exists(os.path.join(self.test_path, 'outdir/output_file1.txt')))
        self.assertEqual(se.exception.code, 1)

    def test_run_sourcerule_succeed(self):
        p=subprocess.Popen(["touch", os.path.join(self.test_path, "outdir/output_file1.txt")])
        p.wait()
        p=subprocess.Popen(["touch", os.path.join(self.test_path, "outdir/output_file2.txt")])
        p.wait()
        p=subprocess.Popen(["touch", os.path.join(self.test_path, "outdir/output_file3.txt")])
        p.wait()
        p=subprocess.Popen(["touch", os.path.join(self.test_path, "outdir/output_file4.txt")])
        p.wait()
        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file1, "-d", self.test_path, "-v", "--since", "rule2"]
        with self.assertRaises(SystemExit) as se:
            WopMars().run(cmd_line)
        self.assertEqual(se.exception.code, 0)

    def test_run_skipping_steps_time_check(self):
        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file2_only_files, "-v"]
        time_unix_ms, time_human = get_current_time()
        start = time_unix_ms
        with self.assertRaises(SystemExit):
           WopMars().run(cmd_line)
        time_unix_ms, time_human = get_current_time()
        end = time_unix_ms
        runtime1 = end - start
        time_unix_ms, time_human = get_current_time()
        start = time_unix_ms
        with self.assertRaises(SystemExit):
           WopMars().run(cmd_line)
        time_unix_ms, time_human = get_current_time()
        end = time_unix_ms
        runtime2 = end - start
        self.assertGreater(runtime1 * 1.5, runtime2)
        # pathlib.Path('outdir/output_file1.txt').unlink()
        PathManager.unlink("outdir/output_file1.txt")

    def test_dry_drun_skipping(self):
        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file2_only_files, "-v"]
        with self.assertRaises(SystemExit):
           WopMars().run(cmd_line)

        with self.assertRaises(SystemExit):
           WopMars().run(cmd_line + ["-n"])

    def test_dry_drun_skipping_all_but_one(self):
        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file1, "-v"]
        with self.assertRaises(SystemExit):
           WopMars().run(cmd_line)
        PathManager.unlink("outdir/output_file7.txt")
        with self.assertRaises(SystemExit):
           WopMars().run(cmd_line + ["-n"])

    def test_run_rerun_runtime(self):

        # Slow run
        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file1, "-v"]
        time_unix_ms, time_human = get_current_time()
        start = time_unix_ms
        with self.assertRaises(SystemExit):
            WopMars().run(cmd_line)
        time_unix_ms, time_human = get_current_time()
        end = time_unix_ms
        runtime1 = end - start

        # Fast run run2<run1
        time_unix_ms, time_human = get_current_time()
        start = time_unix_ms
        with self.assertRaises(SystemExit):
            WopMars().run(cmd_line)
        time_unix_ms, time_human = get_current_time()
        end = time_unix_ms
        runtime2 = end - start
        self.assertGreater(runtime1 * 1.5, runtime2)

        # Middle run: run3>run2
        PathManager.unlink("outdir/output_file1.txt")
        time_unix_ms, time_human = get_current_time()
        start = time_unix_ms
        with self.assertRaises(SystemExit):
            WopMars().run(cmd_line)
        time_unix_ms, time_human = get_current_time()
        end = time_unix_ms
        runtime3 = end - start
        self.assertLess(runtime2, runtime3)

    def get_best_factor(self, full_exec_time, rule_count, maximum_ratio=1):
        average_rule_time = full_exec_time / rule_count
        return 1 + (maximum_ratio * average_rule_time / (1.5 + average_rule_time))

    def test_run_error_in_command_line(self):
        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file1, "--dot", "/usr/"]
        with self.assertRaises(SystemExit) as SE:
            WopMars().run(cmd_line)
        self.assertEqual(SE.exception.code, 2)

        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file1, "--not-known"]
        with self.assertRaises(SystemExit) as SE:
            WopMars().run(cmd_line)
        self.assertEqual(SE.exception.code, 2)

    def test_run_packaged_wrappers(self):
        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file4, "-v", "-d",
                    self.test_path]
        with self.assertRaises(SystemExit) as se:
            WopMars().run(cmd_line)
        self.assertEqual(se.exception.code, 0)

    def test_run_one_tool(self):
        cmd_line = ["python", "tool", "wopmars.tests.resource.wrapper.FooWrapper4",
                  "-i", "{'file': {'input1': 'resource/input_files/input_file1.txt'}}",
                  "-o", "{'file': {'output1': 'outdir/output1.txt'}}", "-D", self.__db_url, "-v", "-d",
                    self.test_path]
        with self.assertRaises(SystemExit) as se:
            WopMars().run(cmd_line)
        self.assertEqual(se.exception.code, 0)

    # AG Do not know why but it is not working
    # def test_core(self):
    #     cmd_line = ["python", "tool", "tests.resource.wrapper.FooWrapperCore",
    #               "-o", "{'table': {'FooBase7': 'tests.resource.model.FooBase7'}}",
    #               "-v", "-D", self.db, "-d", self.test_path]
    #     with self.assertRaises(SystemExit) as se:
    #        WopMars().run(cmd_line)
    #     self.assertEqual(se.exception.code, 0)

    def test_clear_history(self):
        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file1, "-v",
                    "-d", self.test_path]
        cmd_line_clear_history = ["python", "-D", self.__db_url, "-w", self.__example_def_file1, "-v",
                    "-d", self.test_path, "--cleanup-metadata"]
        with self.assertRaises(SystemExit):
            WopMars().run(cmd_line)
        with self.assertRaises(SystemExit):
            WopMars().run(cmd_line_clear_history)
        session = SQLManager.instance().get_session()
        self.assertEqual(session.query(Execution).count(), 0)

    def test_run_target_rule(self):
        # SQLManager.instance().create_all()
        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file1, "-v", "--until", "rule3"]
        with self.assertRaises(SystemExit) as se:
           WopMars().run(cmd_line)
        self.assertEqual(se.exception.code, 0)

    def test_run_input_file_not_ready(self):
        # SQLManager.instance().create_all()
        cmd_line = ["python", "-D", self.__db_url, "-w", self.__example_def_file_input_not_ready, "-v"]
        with self.assertRaises(SystemExit) as se:
           WopMars().run(cmd_line)
        self.assertEqual(se.exception.code, 1)


if __name__ == "__main__":
    unittest.main()



import os
import shutil
import sqlite3
import subprocess
import sys

from unittest import TestCase
from wopmars import WopMars
from wopmars.SQLManager import SQLManager
from wopmars.utils.OptionManager import OptionManager
from wopmars.utils.PathManager import PathManager


class TestExample(TestCase):

    def setUp(self):
        self.test_path = PathManager.get_test_path()  # Get tests path
        OptionManager.initial_test_setup()  # Set tests arguments
        self.db_url = OptionManager.instance()["--database"]
        self.db = self.db_url[10:]
        self.example_dir_path = os.path.join(PathManager.get_package_path(), "data/example")
        self.wopfile = os.path.join(self.example_dir_path, "Wopfile.yml")
        self.working_directory = os.path.join(PathManager.get_package_path(), "data/example")
    #
    def tearDown(self):
        # pass
        # pip.main(['uninstall', 'wopexamplecar', '-y']) # working in travis
        subprocess.run([sys.executable, '-m', 'pip', 'uninstall', 'example', '-y'])
        shutil.rmtree(os.path.join(self.working_directory, "build"), ignore_errors=True)
        shutil.rmtree(os.path.join(self.working_directory, "wopexamplecar.egg-info"), ignore_errors=True)
        PathManager.unlink(self.db)
        OptionManager._drop()
        SQLManager._drop()

    def test_example(self):
        # pip.main(['install', '{}/.'.format(self.example_dir_path), '--upgrade']) # working in travis
        subprocess.run([sys.executable, '-m', 'pip', 'install', '{}/.'.format(self.example_dir_path), '--upgrade'])
        cmd_args = [None, "-v", "--database", self.db_url, "--wopfile", self.wopfile, "--directory", self.working_directory]
        with self.assertRaises(SystemExit) as se:
            WopMars().run(cmd_args)
        self.assertEqual(se.exception.code, 0)

        conn = sqlite3.connect(self.db)
        cursor = conn.execute("SELECT * FROM 'Piece';")
        row_list = sorted(cursor.fetchall())
        self.assertEqual(20, len(row_list))
        conn.close()



import os
import shutil
import unittest
from unittest import TestCase

from wopmars.tests.resource.wrapper.FooWrapper10 import FooWrapper10
from wopmars.tests.resource.wrapper.FooWrapper4 import FooWrapper4
from wopmars.tests.resource.wrapper.FooWrapper5 import FooWrapper5
from wopmars.tests.resource.wrapper.FooWrapper6 import FooWrapper6
from wopmars.tests.resource.wrapper.FooWrapper7 import FooWrapper7
from wopmars.tests.resource.wrapper.FooWrapper8 import FooWrapper8
from wopmars.tests.resource.wrapper.FooWrapper9 import FooWrapper9
from wopmars.tests.resource.wrapper.fooPackage.FooWrapperPackaged import FooWrapperPackaged
from wopmars.SQLManager import SQLManager
from wopmars.models.TableInputOutputInformation import TableInputOutputInformation
from wopmars.models.FileInputOutputInformation import FileInputOutputInformation
from wopmars.models.ToolWrapper import ToolWrapper
from wopmars.models.TypeInputOrOutput import TypeInputOrOutput
from wopmars.Reader import Reader
from wopmars.utils.OptionManager import OptionManager
from wopmars.utils.PathManager import PathManager
from wopmars.utils.SetUtils import SetUtils
from wopmars.utils.WopMarsException import WopMarsException


class TestReader(TestCase):

    def setUp(self):

        OptionManager.initial_test_setup()  # Set tests arguments
        SQLManager.instance().create_all()  # Create database with tables

        session = SQLManager.instance().get_session()
        session.get_or_create(TypeInputOrOutput, defaults={"is_input": True}, is_input=True)
        session.get_or_create(TypeInputOrOutput, defaults={"is_input": False}, is_input=False)
        session.commit()
        self.__session = SQLManager.instance().get_session()
        self.__reader = Reader()

        self.__testdir_path = PathManager.get_test_path()

        # The good -------------------------------:

        self.__example_def_file1_path = os.path.join(self.__testdir_path, "resource/wopfile/example_def_file1.yml")
        self.__example_def_file3_path = os.path.join(self.__testdir_path, "resource/wopfile/example_def_file3.yml")

        # The ugly (malformed file) --------------------:

        self.__s_example_definition_file_duplicate_rule = os.path.join(self.__testdir_path, "resource/wopfile/example_def_file_duplicate_rule.yml")

        self.__list_f_to_exception_init = [
            os.path.join(self.__testdir_path, s_path) for s_path in [
                "resource/wopfile/example_def_file_wrong_yaml.yml",
                "resource/wopfile/example_def_file_duplicate_rule.yml",
                "resource/wopfile/example_def_file_wrong_grammar.yml",
                "resource/wopfile/example_def_file_wrong_grammar2.yml",
                "resource/wopfile/example_def_file_wrong_grammar3.yml",
                "resource/wopfile/example_def_file_wrong_grammar4.yml"
                ]
        ]

        # The bad (invalid file) ----------------------:

        self.__list_s_to_exception_read = [
            os.path.join(self.__testdir_path, s_path) for s_path in [
                "resource/wopfile/example_def_file1.yml",
                "resource/wopfile/example_def_file_wrong_content2.yml",
                "resource/wopfile/example_def_file_wrong_content3.yml",
                "resource/wopfile/example_def_file_wrong_content4.yml",
                "resource/wopfile/example_def_file_wrong_content5.yml",
                "resource/wopfile/example_def_file_wrong_class_name.yml",
                "resource/wopfile/example_def_file_wrong_rule.yml",
            ]
        ]

    # Fix this tests for Travis
    # def test_load_definition_file(self):
    #     for file in self.__list_f_to_exception_init:
    #         with self.assertRaises(WopMarsException):
    #             self.__reader.load_wopfile_as_yml_dic(file)
    #     # Not existing file
    #     with self.assertRaises(WopMarsException):
    #         self.__reader.load_wopfile_as_yml_dic("Not existing file.")

    def test_check_duplicate_rule(self):
        with open(self.__s_example_definition_file_duplicate_rule) as file_duplicate_rule:
            with self.assertRaises(WopMarsException):
                Reader.check_duplicate_rules(file_duplicate_rule.read())

        with open(self.__example_def_file1_path) as file:
            try:
                Reader.check_duplicate_rules(file.read())
            except Exception as e:
                raise AssertionError("Should not raise an exception " + str(e))

    def test_read2(self):
        try:
            self.__reader.iterate_wopfile_yml_dic_and_insert_rules_in_db(self.__example_def_file3_path)
            result = self.__session.query(ToolWrapper).one()
        except:
            raise AssertionError("Packaged wrappers should not raise an exception")

        input_entry = TypeInputOrOutput(is_input=True)
        output_entry = TypeInputOrOutput(is_input=False)

        f1 = FileInputOutputInformation(file_key="input1", path="resource/input_files/input_file1.txt")
        f1.relation_file_or_tableioinfo_to_typeio = input_entry

        f2 = FileInputOutputInformation(file_key="output1", path="outdir/output_file1.txt")
        f2.relation_file_or_tableioinfo_to_typeio = output_entry

        t1 = TableInputOutputInformation(model_py_path="resource.model.fooPackage.FooBasePackaged",
                                         table_key="FooBasePackaged", table_name="FooBasePackaged")
        t1.relation_file_or_tableioinfo_to_typeio = input_entry

        tw1 = FooWrapperPackaged(rule_name="rule1")
        tw1.relation_toolwrapper_to_fileioinfo.extend([f1, f2])
        tw1.relation_toolwrapper_to_tableioinfo.append(t1)

        self.assertEqual(result, tw1)

    def test_read(self):

        self.__reader.iterate_wopfile_yml_dic_and_insert_rules_in_db(self.__example_def_file1_path)
        result = set(self.__session.query(ToolWrapper).all())

        input_entry = TypeInputOrOutput(is_input=True)
        output_entry = TypeInputOrOutput(is_input=False)

        f1 = FileInputOutputInformation(file_key="input1", path="resource/input_files/input_file1.txt")
        f1.relation_file_or_tableioinfo_to_typeio = input_entry

        f2 = FileInputOutputInformation(file_key="output1", path="outdir/output_file1.txt")
        f2.relation_file_or_tableioinfo_to_typeio = output_entry

        f3 = FileInputOutputInformation(file_key="input1", path="outdir/output_file1.txt")
        f3.relation_file_or_tableioinfo_to_typeio = input_entry

        f3bis = FileInputOutputInformation(file_key="input1", path="outdir/output_file1.txt")
        f3bis.relation_file_or_tableioinfo_to_typeio = input_entry

        f4 = FileInputOutputInformation(file_key="output1", path="outdir/output_file2.txt")
        f4.relation_file_or_tableioinfo_to_typeio = output_entry

        f5 = FileInputOutputInformation(file_key="output1", path="outdir/output_file3.txt")
        f5.relation_file_or_tableioinfo_to_typeio = output_entry

        f6 = FileInputOutputInformation(file_key="output2", path="outdir/output_file4.txt")
        f6.relation_file_or_tableioinfo_to_typeio = output_entry

        f7 = FileInputOutputInformation(file_key="input1", path="outdir/output_file3.txt")
        f7.relation_file_or_tableioinfo_to_typeio = input_entry

        f8 = FileInputOutputInformation(file_key="input2", path="outdir/output_file2.txt")
        f8.relation_file_or_tableioinfo_to_typeio = input_entry

        f9 = FileInputOutputInformation(file_key="output1", path="outdir/output_file5.txt")
        f9.relation_file_or_tableioinfo_to_typeio = output_entry

        f10 = FileInputOutputInformation(file_key="input1", path="outdir/output_file4.txt")
        f10.relation_file_or_tableioinfo_to_typeio = input_entry

        f11 = FileInputOutputInformation(file_key="output1", path="outdir/output_file6.txt")
        f11.relation_file_or_tableioinfo_to_typeio = output_entry

        f12 = FileInputOutputInformation(file_key="input1", path="outdir/output_file1.txt")
        f12.relation_file_or_tableioinfo_to_typeio = input_entry

        f13 = FileInputOutputInformation(file_key="input2", path="outdir/output_file5.txt")
        f13.relation_file_or_tableioinfo_to_typeio = input_entry

        f14 = FileInputOutputInformation(file_key="input3", path="outdir/output_file6.txt")
        f14.relation_file_or_tableioinfo_to_typeio = input_entry

        f15 = FileInputOutputInformation(file_key="output1", path="outdir/output_file7.txt")
        f15.relation_file_or_tableioinfo_to_typeio = output_entry

        t1 = TableInputOutputInformation(model_py_path="FooBase", table_key="FooBase", table_name="FooBase")
        t1.relation_file_or_tableioinfo_to_typeio = output_entry

        t1bis = TableInputOutputInformation(model_py_path="FooBase", table_key="FooBase", table_name="FooBase")
        t1bis.relation_file_or_tableioinfo_to_typeio = input_entry

        t2 = TableInputOutputInformation(model_py_path="FooBase2", table_key="FooBase2", table_name="FooBase2")
        t2.relation_file_or_tableioinfo_to_typeio = output_entry

        t2bis = TableInputOutputInformation(model_py_path="FooBase2", table_key="FooBase2", table_name="FooBase2")
        t2bis.relation_file_or_tableioinfo_to_typeio = input_entry

        tw1 = FooWrapper4(rule_name="rule1")
        tw1.relation_toolwrapper_to_fileioinfo.extend([f1, f2])
        tw2 = FooWrapper5(rule_name="rule2")
        tw2.relation_toolwrapper_to_fileioinfo.extend([f3, f4])
        tw2.relation_toolwrapper_to_tableioinfo.extend([t1])
        tw3 = FooWrapper6(rule_name="rule3")
        tw3.relation_toolwrapper_to_fileioinfo.extend([f3bis, f5, f6])
        tw4 = FooWrapper7(rule_name="rule4")
        tw4.relation_toolwrapper_to_tableioinfo.extend([t1bis, t2])
        tw5 = FooWrapper8(rule_name="rule5")
        tw5.relation_toolwrapper_to_fileioinfo.extend([f8, f7, f9])
        tw6 = FooWrapper9(rule_name="rule6")
        tw6.relation_toolwrapper_to_fileioinfo.extend([f10, f11])
        tw6.relation_toolwrapper_to_tableioinfo.extend([t2bis])
        tw7 = FooWrapper10(rule_name="rule7")
        tw7.relation_toolwrapper_to_fileioinfo.extend([f12, f13, f14, f15])

        expected = set([tw1, tw2, tw3, tw4, tw5, tw6, tw7])

        # The good ------------------------------------:
        self.assertTrue((SetUtils.all_elm_of_one_set_in_one_other(result, expected) and
                         SetUtils.all_elm_of_one_set_in_one_other(expected, result)))

        # The bad -------------------------------------:

        # [self.assertRaises(WopMarsException, self.__reader.iterate_wopfile_yml_dic_and_insert_rules_in_db, file) for file in self.__list_s_to_exception_read]

        SQLManager.instance().get_session().rollback()

    def tearDown(self):
        SQLManager.instance().get_session().close()
        SQLManager.instance().drop_all()
        OptionManager._drop()
        SQLManager._drop()
        shutil.rmtree(os.path.join(self.__testdir_path, "outdir"), ignore_errors=True)


if __name__ == "__main__":
    unittest.main()



from wopmars.Base import Base
from sqlalchemy import Column, String, BigInteger, DateTime
from sqlalchemy.orm import relationship
from sqlalchemy.sql.ddl import DDL
from wopmars.SQLManager import SQLManager

class TableModificationTime(Base):
    """
    The TableModificationTime model contains the table names of the workflow and their mtime_epoch_millis of last modification. The table
    ``wom_modification_table`` contains the following fields:

    - table_name: VARCHAR(255) - primary key - the is_input of the table
    - mtime_epoch_millis: INTEGER - unix mtime_epoch_millis [ms] of last modification of the table
    """

    __tablename__ = "wom_{}".format(__qualname__)

    table_name = Column(String(255), primary_key=True)
    mtime_human = Column(DateTime, nullable=False)
    mtime_epoch_millis = Column(BigInteger, nullable=False)

    # One table_io_info has one table_modif_time
    relation_tablemodiftime_to_tableioinfo = relationship("TableInputOutputInformation", back_populates="relation_tableioinfo_to_tablemodiftime")

    def __repr__(self):
        return "<Modification on " + str(self.table_name) + ": " + str(self.mtime_epoch_millis) + ">"

    @classmethod
    def create_triggers(cls):
        """
        Create an INSERT, UPDATE, DELETE trigger on the models created by the user in order to store the modifications mtime_epoch_millis.
        """
        stmt_list = ["INSERT", "UPDATE", "DELETE"]
        for user_table_name in Base.metadata.tables:
            if user_table_name[:4] != "wom_":
                for statement in stmt_list:
                    data={"statement": str(statement), "user_table_name": user_table_name, "wom_table_name": "wom_{}".format(cls.__qualname__)}
                    if SQLManager.instance().__dict__['d_database_config']['db_connection'] == 'sqlite':
                        sql_trigger = "CREATE TRIGGER IF NOT EXISTS {user_table_name}_{statement} " \
                              "AFTER {statement} ON {user_table_name} BEGIN UPDATE {wom_table_name} " \
                              "SET mtime_epoch_millis = CAST((julianday('now') - 2440587.5)*86400000 AS INTEGER), " \
                              "mtime_human = datetime('now', 'localtime') " \
                                      "WHERE table_name = '{user_table_name}'; END;".format(**data)
                    # elif SQLManager.instance().__dict__['d_database_config']['db_connection'] == 'mysql':
                    #     sql_trigger = "CREATE TRIGGER IF NOT EXISTS {table_key}_{statement} AFTER {statement} " \
                    #       "ON {table_key} for each row UPDATE wom_table_modification_time SET " \
                    #                   "mtime_epoch_millis = ROUND(UNIX_TIMESTAMP(CURTIME(4)) * 1000) " \
                    #       "WHERE table_key = '{table_key}';".format(**data)
                    #     obj_ddl = DDL(sql_trigger)
                    #     SQLManager.instance().create_trigger(Base.metadata.tables[table_key], obj_ddl)
                    # elif SQLManager.instance().__dict__['d_database_config']['db_connection'] == 'postgresql':
                    #     sql_trigger = """
                    #         CREATE OR REPLACE FUNCTION {table_key}_{statement}() RETURNS TRIGGER AS ${table_key}_{statement}$
                    #         BEGIN
                    #         UPDATE wom_table_modification_time SET mtime_epoch_millis = extract(epoch from now())*1000 WHERE table_key = '{table_key}';
                    #         RETURN NULL; -- result is ignored since this is an AFTER trigger
                    #         END;
                    #         ${table_key}_{statement}$ LANGUAGE plpgsql;
                    #         DROP TRIGGER IF EXISTS {table_key}_{statement} ON "{table_key}";
                    #         CREATE TRIGGER {table_key}_{statement} AFTER INSERT ON "{table_key}"
                    #         FOR EACH ROW EXECUTE PROCEDURE {table_key}_{statement}();
                    #         """.format(**data)
                    obj_ddl = DDL(sql_trigger)
                    SQLManager.instance().create_trigger(Base.metadata.tables[user_table_name], obj_ddl)



"""
Example of module documentation which can be
multiple-lined
"""
import errno
import importlib
import os
import pathlib
import wopmars


class PathManager:
    """
    Static class for finding paths
    """

    @staticmethod
    def get_package_path():
        """
        Returns the vtam.__path__[0]

        :return: path to the package
        """

        package_path = wopmars.__path__[0]
        return package_path

    @classmethod
    def get_test_path(cls):
        """
        Find the tests output of the project

        :return: the output leading to the tests output of the project
        """

        test_path = os.path.join(cls.get_package_path(), "tests")
        return test_path

    @staticmethod
    def check_pygraphviz(path):
        importlib.import_module("pygraphviz")

    @staticmethod
    def check_database_valid_url(url):
        """
        Check if the path given is correct on the system.

        If path is None, return None.

        :param path:
        :return:
        """
        db_connection = url.split("://")[0]
        if db_connection == "sqlite":
            sqlite_db_path = url.replace("sqlite:///", "")
            if not os.path.isabs(sqlite_db_path):
                sqlite_db_path = os.path.join(os.getcwd(), sqlite_db_path)
            try:
                os.makedirs(os.path.dirname(sqlite_db_path))
            except OSError as exception:
                if exception.errno != errno.EEXIST:
                    raise
            if sqlite_db_path is None or os.access(os.path.dirname(os.path.abspath(os.path.expanduser(sqlite_db_path))), os.W_OK) or sqlite_db_path[0] == "$":
                return sqlite_db_path
            else:
                raise FileNotFoundError

    @staticmethod
    def check_valid_path(path):
        """
        Check if the path given is correct on the system.

        If path is None, return None.

        :param path:
        :return:
        """
        if path is None or os.access(os.path.dirname(os.path.abspath(os.path.expanduser(path))), os.W_OK) or path[0] == "$":
            return path
        else:
            raise FileNotFoundError

    @staticmethod
    def unlink(path):
        """
        Remove a file that may not exist.

        Parameters
        ----------
        path

        Returns
        -------

        """
        try:
            pathlib.Path(path).unlink()
        except FileNotFoundError:
            pass

    @staticmethod
    def dir_content_remove(path):
        for f in os.listdir(path):
            if not f.startswith("."):
                PathManager.unlink(os.path.join(path, f))

    @staticmethod
    def is_in_python_path(name):
        importlib.import_module(name)



"""
Module containing the FooWrapper1 class
"""
import subprocess

from wopmars.models.ToolWrapper import ToolWrapper
from wopmars.utils.Logger import Logger

class FooWrapper5(ToolWrapper):
    """
    This class has been done for example/testing purpose.
    Modifications may lead to failure in tests.
    """
    __mapper_args__ = {'polymorphic_identity': "FooWrapper5"}

    def specify_input_file(self):
        return ["input1"]

    def specify_output_file(self):
        return ["output1"]

    def specify_output_table(self):
        return ["FooBase"]

    def run(self):
        Logger.instance().info(self.__class__.__name__ + " is running...")
        p = subprocess.Popen(["touch", self.output_file("output1")])
        p.wait()
        # self.session.delete_content(self.output_table("FooBase"))
        for i in range(1000):
            # import pdb; pdb.set_trace()
            f = self.output_table("FooBase")(name="Foowrapper5 - {}".format(i))
            self.session.add(f)
        self.session.commit()



"""
Module containing the FooWrapper1 class
"""
import subprocess
import time

from wopmars.utils.Logger import Logger
from wopmars.models.ToolWrapper import ToolWrapper


class FooWrapper10(ToolWrapper):
    """
    This class has been done for example/testing purpose.
    Modifications may lead to failure in tests.
    """
    __mapper_args__ = {'polymorphic_identity': "fooPackage.FooWrapper10"}

    def specify_input_file(self):
        return ["input1", "input2", "input3"]

    def specify_output_file(self):
        return ["output1"]

    def run(self):
        Logger.instance().info(self.__class__.__name__ + " is running...")
        p = subprocess.Popen(["touch", self.output_file("output1")])
        p.wait()
        time.sleep(1)



import shutil
import unittest
from unittest import TestCase

from wopmars.tests.resource.model.FooBase import FooBase
from wopmars.tests.resource.model.FooBase2 import FooBase2
from wopmars.SQLManager import SQLManager
from wopmars.models.TableInputOutputInformation import TableInputOutputInformation
from wopmars.utils.OptionManager import OptionManager
from wopmars.utils.PathManager import PathManager


class TestTableInputOutputInformation(TestCase):

    def setUp(self):

        self.test_path = PathManager.get_test_path()
        OptionManager.initial_test_setup()  # Set tests arguments
        SQLManager.instance().create_all()  # Create database with tables

        self.__local_session = SQLManager.instance().get_session()
        try:
            for i in range(10):
                self.__local_session.add(FooBase(name="testIODB " + str(i)))
            self.__local_session.commit()
        except Exception as e:
            self.__local_session.rollback()
            self.__local_session.close()
            raise e

        self.__io_base_existing = TableInputOutputInformation(model_py_path="FooBase", table_key="FooBase", table_name="FooBase")
        self.__io_base_existing.set_table(FooBase)
        self.__io_base_existing2 = TableInputOutputInformation(model_py_path="FooBase", table_key="FooBase", table_name="FooBase")
        self.__io_base_existing2.set_table(FooBase)
        self.__io_base_existing3 = TableInputOutputInformation(model_py_path="FooBase2", table_key="FooBase2", table_name="FooBase2")
        self.__io_base_existing3.set_table(FooBase2)

    def tearDown(self):
        SQLManager.instance().get_session().close()
        SQLManager.instance().drop_all()
        # PathManager.dir_content_remove(os.path.join(self.test_path, "outdir"))
        shutil.rmtree("outdir", ignore_errors=True)
        OptionManager._drop()
        SQLManager._drop()

    def test_eq(self):
        self.assertEqual(self.__io_base_existing, self.__io_base_existing2)
        self.assertNotEqual(self.__io_base_existing, self.__io_base_existing3)

    def test_is_ready(self):
        self.assertTrue(self.__io_base_existing.is_ready())
        self.assertFalse(self.__io_base_existing3.is_ready())

if __name__ == "__main__":
    unittest.main()



"""
Module containing the FooWrapper1 class
"""
import time

from wopmars.utils.Logger import Logger
from wopmars.models.ToolWrapper import ToolWrapper


class FooWrapper7(ToolWrapper):
    """
    This class has been done for example/testing purpose.
    Modifications may lead to failure in tests.
    """
    __mapper_args__ = {'polymorphic_identity': "fooPackage.FooWrapper7"}
    def specify_input_table(self):
        return ["FooBaseP"]

    def specify_output_table(self):
        return ["FooBase2P"]

    def run(self):
        Logger.instance().info(self.__class__.__name__ + " is running...")
        inputs = self.session.query(self.input_table("FooBaseP")).all()
        self.session.delete_content(self.output_table("FooBase2P"))
        for i in inputs:
            entry = self.output_table("FooBase2P")(name=i.name)
            self.session.add(entry)
        Logger.instance().info(self.session.query(self.output_table("FooBase2P")).all())
        time.sleep(1)



"""
Module containing the FooWrapper1 class
"""
import time
import subprocess
from wopmars.utils.Logger import Logger
from wopmars.models.ToolWrapper import ToolWrapper

class FooWrapper6(ToolWrapper):
    """
    This class has been done for example/testing purpose.
    Modifications may lead to failure in tests.
    """
    __mapper_args__ = {'polymorphic_identity': "FooWrapper6"}

    def specify_input_file(self):
        return ["input1"]

    def specify_output_file(self):
        return ["output1", "output2"]

    def run(self):
        Logger.instance().info(self.__class__.__name__ + " is running...")
        p1 = subprocess.Popen(["touch", self.output_file('output1')])
        p2 = subprocess.Popen(["touch", self.output_file('output2')])
        p1.wait()
        p2.wait()
        time.sleep(0.1)



import codecs
from datetime import datetime
import os
import time


def get_current_time():
    timestamp = time.time()  # epoch mtime_epoch_millis in ms
    timestamp_epoch_millis = timestamp * 1000  # epoch mtime_epoch_millis in ms
    timestamp_human = datetime.fromtimestamp(timestamp)  # human readable local mtime_epoch_millis
    return timestamp_epoch_millis, timestamp_human


def get_mtime(path):
    mtime_epoch = os.path.getmtime(path)  # epoch mtime_epoch_millis in ms
    mtime_epoch_millis = mtime_epoch * 1000  # epoch mtime_epoch_millis in ms
    mtime_human = datetime.fromtimestamp(mtime_epoch)  # mtime in human readable local mtime_epoch_millis
    return mtime_epoch_millis, mtime_human



import os
import shutil
import unittest
from unittest import TestCase

from wopmars.tests.resource.wrapper.FooWrapper2 import FooWrapper2
from wopmars.tests.resource.wrapper.FooWrapper8 import FooWrapper8
from wopmars.SQLManager import SQLManager
from wopmars.models.FileInputOutputInformation import FileInputOutputInformation
from wopmars.models.ToolWrapper import ToolWrapper
from wopmars.models.TypeInputOrOutput import TypeInputOrOutput
from wopmars.DAG import DAG
from wopmars.utils.OptionManager import OptionManager
from wopmars.utils.PathManager import PathManager


class TestDAG(TestCase):

    @classmethod
    def setUpClass(cls):
        shutil.rmtree(os.path.join(PathManager.get_test_path(), 'outdir'), ignore_errors=True)

    def setUp(self):
        OptionManager.initial_test_setup()  # Set tests arguments
        SQLManager.instance().create_all()  # Create database with tables
        #        first
        #       /    \
        #   second   third
        #       \    /
        #       fourth
        #

        self.__session = SQLManager.instance().get_session()

        input_entry = TypeInputOrOutput(is_input=True)
        output_entry = TypeInputOrOutput(is_input=False)

        f1 = FileInputOutputInformation(file_key="input1", path="file1.txt")
        f1.relation_file_or_tableioinfo_to_typeio = input_entry

        f2 = FileInputOutputInformation(file_key="output1", path="file2.txt")
        f2.relation_file_or_tableioinfo_to_typeio = output_entry

        self.__toolwrapper_first = FooWrapper2(rule_name="rule1")
        self.__toolwrapper_first.relation_toolwrapper_to_fileioinfo.extend([f1, f2])

        f1 = FileInputOutputInformation(file_key="input1", path="file2.txt")
        f1.relation_file_or_tableioinfo_to_typeio = input_entry

        f2 = FileInputOutputInformation(file_key="output1", path="file3.txt")
        f2.relation_file_or_tableioinfo_to_typeio = output_entry

        self.__toolwrapper_second = FooWrapper2(rule_name="rule2")
        self.__toolwrapper_second.relation_toolwrapper_to_fileioinfo.extend([f1, f2])

        f1 = FileInputOutputInformation(file_key="input1", path="file2.txt")
        f1.relation_file_or_tableioinfo_to_typeio = input_entry

        f2 = FileInputOutputInformation(file_key="output1", path="file4.txt")
        f2.relation_file_or_tableioinfo_to_typeio = output_entry

        self.__toolwrapper_third = FooWrapper2(rule_name="rule3")
        self.__toolwrapper_third.relation_toolwrapper_to_fileioinfo.extend([f1, f2])

        f1 = FileInputOutputInformation(file_key="input1", path="file3.txt")
        f1.relation_file_or_tableioinfo_to_typeio = input_entry

        f2 = FileInputOutputInformation(file_key="input2", path="file4.txt")
        f2.relation_file_or_tableioinfo_to_typeio = input_entry

        f3 = FileInputOutputInformation(file_key="output1", path="file5.txt")
        f3.relation_file_or_tableioinfo_to_typeio = output_entry

        self.__toolwrapper_fourth = FooWrapper8(rule_name="rule4")
        self.__toolwrapper_fourth.relation_toolwrapper_to_fileioinfo.extend([f1, f2, f3])

        list_tool = [self.__toolwrapper_first,
                     self.__toolwrapper_second,
                     self.__toolwrapper_third,
                     self.__toolwrapper_fourth]

        self.__set_tool = set(list_tool)

        SQLManager.instance().get_session().add_all(list_tool)
        SQLManager.instance().get_session().commit()

    def test_get_all_successors(self):
        my_dag = DAG(self.__set_tool)

        self.assertEqual(set(my_dag.get_all_successors(self.__toolwrapper_first)), self.__set_tool)
        self.assertNotEqual(set(my_dag.get_all_successors(self.__toolwrapper_first)),
                            self.__set_tool.difference(set([self.__toolwrapper_second])))
        self.assertNotEqual(set(my_dag.get_all_successors(self.__toolwrapper_first)),
                            self.__set_tool.difference(set([self.__toolwrapper_first])))

    def test_get_all_predecessors(self):
        my_dag = DAG(self.__set_tool)

        self.assertEqual(set(my_dag.get_all_predecessors(self.__toolwrapper_fourth)), self.__set_tool)
        self.assertNotEqual(set(my_dag.get_all_predecessors(self.__toolwrapper_fourth)),
                            self.__set_tool.difference(set([self.__toolwrapper_second])))
        self.assertNotEqual(set(my_dag.get_all_predecessors(self.__toolwrapper_fourth)),
                            self.__set_tool.difference(set([self.__toolwrapper_fourth])))

    def tearDown(self):
        SQLManager.instance().get_session().close() 
        SQLManager.instance().drop_all()
        OptionManager._drop()
        PathManager.unlink("tests/outdir/output_file1.txt")
        SQLManager._drop()

    def test_init(self):
        try:
            my_dag = DAG(self.__set_tool)

            dag_from_base = DAG(set(SQLManager.instance().get_session().query(ToolWrapper).all()))
            self.assertEqual(my_dag, dag_from_base)

            # Verifying that the nodes are correctly sorted
            self.assertTrue(self.__toolwrapper_fourth in my_dag.successors(self.__toolwrapper_third))
            self.assertTrue(self.__toolwrapper_fourth in my_dag.successors(self.__toolwrapper_second))
            self.assertTrue(self.__toolwrapper_second in my_dag.successors(self.__toolwrapper_first))
            self.assertTrue(self.__toolwrapper_third in my_dag.successors(self.__toolwrapper_first))
        except:
            # beware, if a test_bak inside the try block fails, 2 exceptions are raised
            raise AssertionError('Should not raise exception')

if __name__ == '__main__':
    unittest.main()




import os
import shutil
import unittest
from unittest import TestCase

from wopmars.SQLManager import SQLManager
from wopmars.models.FileInputOutputInformation import FileInputOutputInformation
from wopmars.utils.OptionManager import OptionManager
from wopmars.utils.PathManager import PathManager


class TestFileInputOutputInformation(TestCase):
    def setUp(self):
        self.test_path = PathManager.get_test_path()  # Get tests path
        OptionManager.initial_test_setup()  # Set tests arguments

        s_path_to_example_existing_file = os.path.join(self.test_path, "resource/input_files/example_existing_file.txt")
        s_path_to_example_existing_file2 = os.path.join(self.test_path, "resource/input_files/example_existing_file2.txt")
        s_path_to_example_not_existing_file = os.path.join(self.test_path, "resource/input_files/example_not_existing_file.txt")

        self.__io_file_existing = FileInputOutputInformation(file_key="existing_file", path=s_path_to_example_existing_file)
        self.__io_file_existing2 = FileInputOutputInformation(file_key="existing_file", path=s_path_to_example_existing_file)
        self.__io_file_existing3 = FileInputOutputInformation(file_key="existing_file2", path=s_path_to_example_existing_file)
        self.__io_file_existing4 = FileInputOutputInformation(file_key="existing_file", path=s_path_to_example_existing_file2)
        self.__io_file_not_existing = FileInputOutputInformation(file_key="not_existing_file", path=s_path_to_example_not_existing_file)

    def test_is_ready(self):
        self.assertTrue(self.__io_file_existing.is_ready())
        self.assertFalse(self.__io_file_not_existing.is_ready())

    def test_eq(self):
        self.assertEqual(self.__io_file_existing, self.__io_file_existing2)
        self.assertNotEqual(self.__io_file_existing, self.__io_file_existing3)
        self.assertNotEqual(self.__io_file_existing, self.__io_file_existing4)
        self.assertNotEqual(self.__io_file_existing, self.__io_file_not_existing)

    def tearDown(self):
        SQLManager.instance().get_session().close()
        SQLManager.instance().drop_all()
        shutil.rmtree(os.path.join(self.test_path, "outdir"), ignore_errors=True)
        OptionManager._drop()
        SQLManager._drop()

if __name__ == "__main__":
    unittest.main()



__version__ = "0.1.4"

import os
import pathlib
import re
import shutil
import sys

from docopt import docopt, DocoptExit
from schema import Schema, And, Or, Use, SchemaError

from wopmars.SQLManager import SQLManager
from wopmars.WorkflowManager import WorkflowManager
from wopmars.utils.DictUtils import DictUtils
from wopmars.utils.Logger import Logger
from wopmars.utils.OptionManager import OptionManager
from wopmars.utils.PathManager import PathManager
from wopmars.utils.WopMarsException import WopMarsException
from wopmars.utils.various import get_mtime, get_current_time


__doc__ = """wopmars %s

Usage:
  wopmars --version | (-D DATABASE) (-w DEFINITION_FILE) [-n] [-F] [-v...] [-d DIR] [-g FILE] [-L FILE] [-S RULE | -U RULE] [-c] [-t]
  wopmars tool TOOLWRAPPER [-i DICT] [-o DICT] [-P DICT] [-F] [-D DATABASE] [-v...] [-d DIR] [-L FILE] [-g FILE] [-c] [-t]
  wopmars example [-d DIR]

Arguments:
  DATABASE         Path to the sqlite database file (Required)
  DEFINITION_FILE  Path to the definition file of the workflow (Required)
  FILE             Path to a file.
  DIR              Path to a directory.
  RULE             Name of a rule in the workflow definition file.
  TOOLWRAPPER      Path the the tool_python_path
  DICT             String formatted like a dictionary. Ex: "{'input1': 'path/to/input1', 'input2': 'path/to/input2'}"

Options:
  --version                    Show version and exists
  -D --database=DATABASE       REQUIRED: Set the path to the database, e.g -D sqlite:///db.sqlite
  -F --forceall                Force the execution of the workflow, without checking for previous executions.
  -L FILE --log=FILE           Write logs in FILE file.
  -P --params=DICT             Set the parameters of the tool_python_path you want to use in the dictionary format.
  -S RULE --since=RULE         Execute the workflow since the given RULE.
  -U RULE --until=RULE         Execute the workflow until the given RULE.
  -c --cleanup-metadata        Clear WoMars history. Should be used in case of bug which seem to be related to the history. Be carefull, clearing history will result in a re-execution of the whole workflow.
  -d --directory=DIR           Specify working directory (relative paths in the wopfile will use this as their origin). [default: $CWD].
  -g FILE --dot=FILE           Write dot representing the workflow in the FILE file (with .dot extension). This option needs to install WopMars with pygraphviz (pip install wopmars[pygraphviz])
  -h --help                    Show this help.
  -i --input=DICT              Set the input of the tool_python_path you want to use in the dictionary format.
  -n --dry-run                 Only display what would have been done.
  -o --output=DICT             Set the output of the tool_python_path you want to use in the dictionary format.
  -t --touch                   Only display what would have been done.
  -u --update                  Should be used when a file supposedly generated by the workflow already exists and should be used as it. (Not implemented)
  -v                           Set verbosity level, eg -v, -vv or -vvv
  -w --wopfile=DEFINITION_FILE REQUIRED: Set the path to the definition file.

Example:
    wopmars example
    cd example
    pip install .
    wopmars -w Wopfile.yml -D "sqlite:///db.sqlite" -v
    
Description:
    WopMars - Workflow Python Manager for Reproducible Science
    Gonzalez A., Giffon L., Spinelli L. 2016-2020
    https://wopmars.readthedocs.io
    https://github.com/aitgon/wopmars
"""%__version__


class WopMars:

    @staticmethod
    def run(argv):

        """
        Entry-point of the program
        """

        # if the command line is malformed, docopt interrupt the software.
        try:
            if argv[1:] == []: # If not arguments, run the help
                argv.append('-h')
            OptionManager.instance().update(docopt(__doc__, argv=argv[1:]))

        except DocoptExit as SE:
            print("Bad argument in the command line: \n\t" + " ".join(argv) + "\n" + str(SE))
            sys.exit(2)

        try:
            schema_option = Schema({
                '--wopfile': Or("Wopfile.yml", str),
                '--database': Use(PathManager.check_database_valid_url),
                '-v': Or(0, And(int, lambda n: 1 <= n <= 2)),
                '--dot': Or(None, And(Use(PathManager.check_valid_path), Use(PathManager.check_pygraphviz))),
                "--log": Use(PathManager.check_valid_path),
                # '--printtools': Use(bool),
                "--since": Or(None, str),
                "--until": Or(None, str),
                "--forceall": Use(bool),
                "--dry-run": Use(bool),
                "--touch": Use(bool),
                "--directory": Use(lambda path: pathlib.Path(path).mkdir(parents=True, exist_ok=True)),
                "--input": Use(DictUtils.str_to_dict),
                "--output": Use(DictUtils.str_to_dict),
                "--params": Use(DictUtils.str_to_dict),
                "TOOLWRAPPER": Or(None, Use(PathManager.is_in_python_path)),
                "tool": Use(bool),
                "example": Use(bool),
                "--version": Use(bool),
                "--cleanup-metadata": Use(bool),
            })
            # The option values are validated using schema library
            OptionManager.instance().validate(schema_option)
            os.chdir(OptionManager.instance()["--directory"])

        except SchemaError as schema_msg:
            Logger.instance().debug("\nCommand line Args:" + str(OptionManager.instance()))
            # regex for the different possible error messages.
            match_open_def = re.match(r"^open\('(.[^\)]+)'\)", str(schema_msg))
            match_dot_def = re.match(r"^check_valid_path\(('.[^\)]+')\)", str(schema_msg))
            match_wrong_key = re.match(r"^Wrong keys ('.[^\)]+')", str(schema_msg))
            match_pygraphviz = re.match(r".*dot.*", str(schema_msg))
            print(match_pygraphviz)
            # Check the different regex..
            if match_open_def:
                Logger.instance().error("The file " + match_open_def.group(1) + " cannot be opened. It may not exist.")
            elif match_dot_def:
                Logger.instance().error("The path " + match_dot_def.group(1) + " is not valid.")
            elif match_wrong_key:
                # Normally never reach
                Logger.instance().error("The option key " + match_wrong_key.group(1) + " is not known.")
            elif match_pygraphviz:
                Logger.instance().error("The dot file path is not valid or the pygraphviz module is not installed. In the second case, install wopmars with pygraphviz: pip install wopmars[pygraphviz]")
            else:
                # Normally never reach
                Logger.instance().error("An unknown error has occured. Message: " + str(schema_msg))
            sys.exit(2)

        Logger.instance().debug("\nCommand line Args:" + str(OptionManager.instance()))

        ############################################################################################
        #
        # Print version to stdout and exists
        #
        ############################################################################################

        if OptionManager.instance()["--version"]:
            print("wopmars {}".format(__version__), file=sys.stdout)
            sys.exit(0)

        ############################################################################################
        #
        # Recursively writes quickstart example and exists
        #
        ############################################################################################

        if OptionManager.instance()["example"]:
            # ExampleBuilder().build()

            source_path = os.path.join(PathManager.get_package_path(), "data/example")
            destination_path = os.path.join("example")

            shutil.rmtree(destination_path, ignore_errors=True)
            shutil.copytree(source_path, destination_path)

            sys.exit(0)

        ############################################################################################
        #
        # Initiates new WorkflowManager instance
        #
        ############################################################################################

        workflow_manager = WorkflowManager()

        ############################################################################################
        #
        # Cleans up non fully terminated executions
        #
        ############################################################################################

        SQLManager.instance().clean_up_unexecuted_tool_wrappers()

        ############################################################################################
        #
        # --cleanup-metadata (clear history and exit)
        #
        ############################################################################################

        if OptionManager.instance()["--cleanup-metadata"]:
            Logger.instance().info("Deleting Wopmars history...")
            # Check if sqlite db path exists
            if pathlib.Path(SQLManager.instance().d_database_config['db_database']).is_file():
                SQLManager.instance().clear_wopmars_history()
            if OptionManager.instance()["--cleanup-metadata"]:
                sys.exit(0)

        try:
            workflow_manager.run()
        except WopMarsException as WE:
            Logger.instance().error(str(WE))
            try:
                timestamp_epoch_millis, timestamp_human = get_current_time()
                Logger.instance().error("The workflow has encountered an error at: {}".format(timestamp_human))
                workflow_manager.set_finishing_informations(timestamp_human, "ERROR")
            except AttributeError:
                SQLManager.instance().get_session().rollback()
                Logger.instance().error("The execution has not even begun. No informations will be stored in the database.")
            except Exception as e:
                Logger.instance().error("An error occurred during the rollback of the changement of the database which can be now unstable:" +
                                        str(e))
            sys.exit(1)
        except Exception as e:
            Logger.instance().error("An unknown error has occurred:\n" + str(e))
            sys.exit(1)


def run():

    sys.path.append(os.path.dirname(os.path.realpath(__file__)) + "/toolwrappers/")
    sys.path.append(os.path.dirname(os.path.realpath(__file__)) + "/base/")

    WopMars().run(sys.argv)

if __name__ == "__main__":
    run()



from sqlalchemy import Column, Integer, String, Boolean
from sqlalchemy.orm import relationship

from wopmars.Base import Base


class TypeInputOrOutput(Base):
    """
    This class is the model of the table ``wom_type_input_or_output``. It stores the two kind of entry named "input" and "output" which
    are associated with an ID.

    Fields:

    - id: INTEGER - primary key -arbitrary ID
    - is_input: VARCHAR(255) - the is_input of the type ("input" or "output")
    """

    __tablename__ = "wom_{}".format(__qualname__)

    is_input = Column(Boolean, primary_key=True, autoincrement=False)

    # One type_io is used by many tables
    relation_typeio_to_tableioinfo = relationship("TableInputOutputInformation", back_populates="relation_file_or_tableioinfo_to_typeio")
    # One type_io is used by many files
    relation_typeio_to_fileioinfo = relationship("FileInputOutputInformation", back_populates="relation_file_or_tableioinfo_to_typeio")

    def __repr__(self):
        return "<class TypeInputOrOutput: is_input {}>".format(self.is_input)



import importlib

from sqlalchemy.exc import OperationalError

from wopmars.Base import Base
from sqlalchemy import Column, Integer, String, ForeignKey, BigInteger, Boolean, DateTime
from sqlalchemy.orm import relationship, reconstructor

from wopmars.SQLManager import SQLManager
from wopmars.models.InputOutput import InputOutput
from wopmars.models.ToolWrapper import ToolWrapper
from wopmars.utils.Logger import Logger
from sqlalchemy.sql.functions import func


class TableInputOutputInformation(InputOutput, Base):
    """
    This class extends InputOutput and is specific to the input or output tables. It is the model which store the references
    to the actual tables needed by the user. The table ``wom_table_io_information`` associated with this model contains the
    following fields:

    - id: INTEGER - primary key - autoincrement - arbitrary ID
    - tablename: VARCHAR(255) - foreign key to the associated table: :class:`wopmars.framework.database.relation_toolwrapper_to_tableioinfo.TableModificationTime.TableModificationTime` - the is_input of the referenced table
    - model: VARCHAR(255) - the path to the model (in python notation)
    - toolwrapper_id: INTEGER - foreign key to the associated rule ID: :class:`wopmars.framework.database.relation_toolwrapper_to_tableioinfo.ToolWrapper.ToolWrapper`
    - is_input: INTEGER - foreign key to the associated type ID: :class:`wopmars.framework.database.relation_toolwrapper_to_tableioinfo.TypeInputOrOutput.TypeInputOrOutput`
    - mtime_epoch_millis: INTEGER - unix mtime_epoch_millis at which the table have been used
    """

    __tablename__ = "wom_{}".format(__qualname__)

    id = Column(Integer, primary_key=True, autoincrement=True)
    table_key = Column(String(255))
    table_name = Column(String(255), ForeignKey("wom_TableModificationTime.table_name"))
    model_py_path = Column(String(255))
    toolwrapper_id = Column(Integer, ForeignKey("wom_ToolWrapper.id", ondelete='CASCADE'))
    is_input = Column(Boolean, ForeignKey("wom_TypeInputOrOutput.is_input"))
    mtime_human = Column(DateTime, nullable=True)
    mtime_epoch_millis = Column(BigInteger, nullable=True)

    # One toolwrapper uses many tables
    relation_file_or_tableioinfo_to_toolwrapper = relationship("ToolWrapper", back_populates="relation_toolwrapper_to_tableioinfo", enable_typechecks=False)
    # One type_io is used by many tables
    relation_file_or_tableioinfo_to_typeio = relationship("TypeInputOrOutput", back_populates="relation_typeio_to_tableioinfo")
    # One table_io_info has one table_modif_time
    relation_tableioinfo_to_tablemodiftime = relationship("TableModificationTime", back_populates="relation_tablemodiftime_to_tableioinfo")

    # all the model names met since the begining of this instance of WopMaRS
    tablemodelnames = set()
    # al the table names met since the begining of this instance of WopMaRS
    tablenames = set()

    def __init__(self, model_py_path, table_key, table_name):
        """
        self.__table is initialized to None and will contain the model of this TableInputOutputInformation object.

        :param model_py_path: The path to the model
        :type model_py_path: str
        :param table_key: The is_input of the table associated with the model
        :type table_key: str
        """
        # The file containing the table should be in PYTHONPATH
        Base.__init__(self, model_py_path=model_py_path, table_key=table_key, table_name=table_name)
        Logger.instance().debug(str(model_py_path) + " model_py_path loaded. Tablename: " + str(table_key))
        self.__table = None

    @reconstructor
    def init_on_load(self):
        """
        This is used by SQLAlchemy to regenerate the right object when loading it from the database. Here, we need to
        get back the actual Model from the model is_input and store it in self.__table.
        """
        for table in TableInputOutputInformation.tablemodelnames:
            mod = importlib.import_module(table)
            try:
                if table == self.model_py_path:
                    # toodo LucG tabling
                    self.__table = eval("mod." + self.model_py_path.split(".")[-1])
            except AttributeError as e:
                raise e
        Logger.instance().debug(self.table_key + " table class reloaded. Model: " + self.model_py_path)

    def set_table(self, model):
        self.__table = model

    def get_table(self):
        return self.__table

    @staticmethod
    def set_tables_properties(tables):
        """
        Import the models of the current execution and then associate models with TableInputOutputInformation objects.

        :param tables: the TableInputOutputInformation which need their table properties to be set.
        :type tables: ResultSet(TableInputOutputInformation)
        """
        # import models for avoid references errors between models when dealing with them
        TableInputOutputInformation.import_models(set([t.model_py_path for t in tables]))

        for table in tables:
            # keep track of the models used in static variable of TableInputOutputInformation
            TableInputOutputInformation.tablemodelnames.add(table.model_py_path)
            # Associate model with the TableInputOutputInformation object
            mod = importlib.import_module(table.model_py_path)
            table_model = eval("mod." + table.model_py_path.split(".")[-1])
            table.set_table(table_model)
            # keep track of table names used in static variable of TableInputOutputInformation
            TableInputOutputInformation.tablenames.add(table_model.__tablename__)
            SQLManager.instance().get_session().add(table)

    @staticmethod
    def get_execution_tables():
        """
        Return all the TableInputOutputInformation objects found in model TableInputOutputInformation.

        :return: ResultSet TableInputOutputInformation objects
        """
        session = SQLManager.instance().get_session()
        execution_id = session.query(func.max(ToolWrapper.execution_id))
        return session.query(TableInputOutputInformation).filter(TableInputOutputInformation.toolwrapper_id == ToolWrapper.id).filter(ToolWrapper.execution_id == execution_id).all()

    @staticmethod
    def import_models(model_names):
        """
        Import all the given models

        :param model_names: The path to the models
        :type model_names: Iterable(String)
        """
        for t in model_names:
            Logger.instance().debug("TableInputOutputInformation.import_models: importing " + str(t))
            importlib.import_module(t)

    def is_ready(self):
        """
        A TableInputOutputInformation object is ready if its table exists and contains entries.

        :return: bool if the table is ready
        """
        session = SQLManager.instance().get_session()
        try:
            results = session.query(self.__table).first()
            if results is None:
                Logger.instance().debug("The table " + self.table_key + " is empty.")
                return False
        except OperationalError as e:
            Logger.instance().debug("The table " + self.__table.__tablename__ + " doesn't exist.")
            return False
        except Exception as e:
            session.rollback()
            raise e
            # toodo LG twthread
        return True

    def __eq__(self, other):
        """
        Two TableInputOutputInformation object are equals if their table attributes belongs to the same class and if the associated table
        has the same content

        :param other: TableInputOutputInformation
        :return: boolean: True if the table attributes are the same, False if not
        """
        session = SQLManager.instance().get_session()
        if self.model_py_path != other.model_py_path or self.table_key != other.table_key:
            return False
        try:
            self_results = set(session.query(self.__table).all())
            other_results = set(session.query(other.get_table()).all())
            if self_results != other_results:
                return False
        except Exception as e:
            session.rollback()
            raise e
        return True

    def __hash__(self):
        return id(self)

    def __repr__(self):
        return "<class {}; tablename: {}; model_py_path: {}; used_at: {}>"\
            .format(self.__class__.__name__, self.table_key, self.model_py_path, self.mtime_epoch_millis)

    def __str__(self):
        return "<Table: " + self.table_key + "; model: " + self.model_py_path + ">"



import shutil
import time
import unittest
from unittest import TestCase

from wopmars.tests.resource.model.FooBase import FooBase
from wopmars.tests.resource.wrapper.FooWrapper5 import FooWrapper5
from wopmars.tests.resource.wrapper.sprintFive.Add import Add
from wopmars.tests.resource.wrapper.sprintFive.Query import Query
from wopmars.SQLManager import SQLManager
from wopmars.models.TableInputOutputInformation import TableInputOutputInformation
from wopmars.models.FileInputOutputInformation import FileInputOutputInformation
from wopmars.models.TableModificationTime import TableModificationTime
from wopmars.models.Option import Option
from wopmars.models.TypeInputOrOutput import TypeInputOrOutput
from wopmars.ToolWrapperThread import ToolWrapperThread
from wopmars.utils.OptionManager import OptionManager
from wopmars.utils.PathManager import PathManager
from wopmars.utils.various import get_current_time


class TestToolWrapperThread(TestCase):

    def setUp(self):
        OptionManager.initial_test_setup()  # Set tests arguments
        SQLManager.instance().create_all()  # Create database with tables

        [SQLManager.instance().get_session().add(FooBase(name="foo " + str(i))) for i in range(10000)]
        SQLManager.instance().get_session().commit()

    def test_run(self):

        input_entry = TypeInputOrOutput(is_input=True)
        output_entry = TypeInputOrOutput(is_input=False)

        f1 = FileInputOutputInformation(file_key="input1", path="resource/input_files/input_file1.txt")
        f1.relation_file_or_tableioinfo_to_typeio = input_entry

        f2 = FileInputOutputInformation(file_key="output1", path="outdir/output_file1.txt")
        f2.relation_file_or_tableioinfo_to_typeio = output_entry

        t1 = TableInputOutputInformation(model_py_path="FooBase", table_key="FooBase", table_name="FooBase")
        t1.set_table(FooBase)
        t1.relation_file_or_tableioinfo_to_typeio = output_entry
        timestamp_millis, timestamp_human = get_current_time()
        modification_table_entry = TableModificationTime(mtime_epoch_millis=timestamp_millis, table_name=t1.table_name)
        t1.modification = modification_table_entry

        tw1 = FooWrapper5(rule_name="rule1")
        tw1.relation_toolwrapper_to_fileioinfo.extend([f1, f2])
        tw1.relation_toolwrapper_to_tableioinfo.append(t1)

        f12 = FileInputOutputInformation(file_key="input1", path="resource/input_files/input_file1.txt")
        f12.relation_file_or_tableioinfo_to_typeio = input_entry

        f22 = FileInputOutputInformation(file_key="output1", path="outdir/output_file1.txt")
        f22.relation_file_or_tableioinfo_to_typeio = output_entry

        t12 = TableInputOutputInformation(model_py_path="FooBase", table_key="FooBase", table_name="FooBase")
        t12.set_table(FooBase)
        t12.relation_file_or_tableioinfo_to_typeio = output_entry
        timestamp_millis, timestamp_human = get_current_time()
        modification_table_entry = TableModificationTime(
            mtime_epoch_millis=timestamp_millis, table_name=t12.table_name)
        t12.modification = modification_table_entry

        tw2 = FooWrapper5(rule_name="rule2")
        tw2.relation_toolwrapper_to_fileioinfo.extend([f12, f22])
        tw2.relation_toolwrapper_to_tableioinfo.append(t12)

        f13 = FileInputOutputInformation(file_key="input1", path="resource/input_files/input_file1.txt")
        f13.relation_file_or_tableioinfo_to_typeio = input_entry

        f23 = FileInputOutputInformation(file_key="output1", path="outdir/output_file1.txt")
        f23.relation_file_or_tableioinfo_to_typeio = output_entry

        t13 = TableInputOutputInformation(model_py_path="FooBase", table_key="FooBase", table_name="FooBase")
        t13.set_table(FooBase)
        t13.relation_file_or_tableioinfo_to_typeio = output_entry
        timestamp_millis, timestamp_human = get_current_time()
        modification_table_entry = TableModificationTime(
            mtime_epoch_millis=timestamp_millis, table_name=t13.table_name)
        t13.modification = modification_table_entry

        tw3 = FooWrapper5(rule_name="rule3")
        tw3.relation_toolwrapper_to_fileioinfo.extend([f13, f23])
        tw3.relation_toolwrapper_to_tableioinfo.append(t13)

        tt1 = ToolWrapperThread(tw1)
        tt2 = ToolWrapperThread(tw2)
        tt3 = ToolWrapperThread(tw3)

        tt1.start()
        tt2.start()
        tt3.start()

        tt1.join()
        tt2.join()
        tt3.join()

        self.assertEqual(len(SQLManager.instance().get_session().query(FooBase).filter(FooBase.name.like('Foowrapper5 - %')).all()), 3000)

    def test_run_commit_vs_query(self):
        # this tests does not work with mysql and postgresql
        if not SQLManager.instance().engine.url.drivername in ['mysql', 'postgresql']:
            input_entry = TypeInputOrOutput(is_input=True)
            output_entry = TypeInputOrOutput(is_input=False)

            f1 = FileInputOutputInformation(file_key="input1", path="resource/input_files/input_file1.txt")
            f1.relation_file_or_tableioinfo_to_typeio = input_entry

            t1 = TableInputOutputInformation(model_py_path="FooBase", table_key="FooBase", table_name="FooBase")
            t1.set_table(FooBase)
            t1.relation_file_or_tableioinfo_to_typeio = output_entry
            timestamp_millis, timestamp_human = get_current_time()
            modification_table_entry = TableModificationTime(mtime_epoch_millis=timestamp_millis, table_name=t1.table_name)
            t1.modification = modification_table_entry

            o1 = Option(name="rows", value="1000")

            tw1 = Add(rule_name="rule1")
            tw1.relation_toolwrapper_to_fileioinfo.append(f1)
            tw1.relation_toolwrapper_to_tableioinfo.append(t1)
            tw1.relation_toolwrapper_to_option.append(o1)

            f12 = FileInputOutputInformation(file_key="input1", path="resource/input_files/input_file1.txt")
            f12.relation_file_or_tableioinfo_to_typeio = input_entry

            t12 = TableInputOutputInformation(model_py_path="FooBase", table_key="FooBase", table_name="FooBase")
            t12.set_table(FooBase)
            t12.relation_file_or_tableioinfo_to_typeio = output_entry
            timestamp_millis, timestamp_human = get_current_time()
            modification_table_entry = TableModificationTime(mtime_epoch_millis=timestamp_millis,
                                                             table_name=t12.table_name)
            t12.modification = modification_table_entry

            o12 = Option(name="rows", value="1000")

            tw12 = Add(rule_name="rule1")
            tw12.relation_toolwrapper_to_fileioinfo.append(f12)
            tw12.relation_toolwrapper_to_tableioinfo.append(t12)
            tw12.relation_toolwrapper_to_option.append(o12)

            f13 = FileInputOutputInformation(file_key="input1", path="resource/input_files/input_file1.txt")
            f13.relation_file_or_tableioinfo_to_typeio = input_entry

            t13 = TableInputOutputInformation(model_py_path="FooBase", table_key="FooBase", table_name="FooBase")
            t13.set_table(FooBase)
            t13.relation_file_or_tableioinfo_to_typeio = output_entry
            timestamp_millis, timestamp_human = get_current_time()
            modification_table_entry = TableModificationTime(mtime_epoch_millis=timestamp_millis,
                                                             table_name=t13.table_name)
            t13.modification = modification_table_entry

            o13 = Option(name="rows", value="1000")

            tw13 = Add(rule_name="rule1")
            tw13.relation_toolwrapper_to_fileioinfo.append(f13)
            tw13.relation_toolwrapper_to_tableioinfo.append(t13)
            tw13.relation_toolwrapper_to_option.append(o13)

            tt1 = ToolWrapperThread(tw1)
            tt2 = ToolWrapperThread(tw12)
            tt3 = ToolWrapperThread(tw13)

            t21 = TableInputOutputInformation(model_py_path="FooBase", table_key="FooBase", table_name="FooBase")
            t21.set_table(FooBase)
            t21.relation_file_or_tableioinfo_to_typeio = input_entry

            tw21 = Query(rule_name="rule1")
            tw21.relation_toolwrapper_to_tableioinfo.append(t21)

            t22 = TableInputOutputInformation(model_py_path="FooBase", table_key="FooBase", table_name="FooBase")
            t22.set_table(FooBase)
            t22.relation_file_or_tableioinfo_to_typeio = input_entry

            tw22 = Query(rule_name="rule1")
            tw22.relation_toolwrapper_to_tableioinfo.append(t22)

            t23 = TableInputOutputInformation(model_py_path="FooBase", table_key="FooBase", table_name="FooBase")
            t23.set_table(FooBase)
            t23.relation_file_or_tableioinfo_to_typeio = input_entry

            tw23 = Query(rule_name="rule1")
            tw23.relation_toolwrapper_to_tableioinfo.append(t23)

            tt4 = ToolWrapperThread(tw21)
            tt5 = ToolWrapperThread(tw22)
            tt6 = ToolWrapperThread(tw23)

            tt4.start()

            tt1.start()
            tt2.start()
            tt3.start()
            time.sleep(5)
            tt5.start()
            tt6.start()

            tt1.join()
            tt2.join()
            tt3.join()
            tt4.join()
            tt5.join()
            tt6.join()

    def tearDown(self):
        SQLManager.instance().get_session().close()
        SQLManager.instance().drop_all()
        shutil.rmtree("wopmars/tests/outdir", ignore_errors=True)

if __name__ == '__main__':
    unittest.main()



import os
import pathlib
import shlex
import shutil
import subprocess
import sys
import unittest

from wopmars.utils.PathManager import PathManager


class TestCommandExample(unittest.TestCase):

    """Will test 'wopmars example'"""

    def setUp(self):

        self.test_path = PathManager.get_test_path()
        self.outdir_path = os.path.join(self.test_path, 'outdir')
        pathlib.Path(self.outdir_path).mkdir(exist_ok=True, parents=True)

    def test_command_example(self):

        cmd = "wopmars example"

        if sys.platform.startswith("win"):
            args = cmd
        else:
            args = shlex.split(cmd)

        result = subprocess.run(args=args, check=True, cwd=self.outdir_path)
        self.assertEqual(result.returncode, 0)

    def tearDown(self):
        shutil.rmtree(self.outdir_path, ignore_errors=True)



"""
Module containing the Logger class.
"""
import logging
import sys
from logging.handlers import RotatingFileHandler

from wopmars.utils.OptionManager import OptionManager
from wopmars.utils.Singleton import SingletonMixin
from wopmars.utils.ColorPrint import ColorPrint


class LessThanFilter(logging.Filter):
    def __init__(self, exclusive_maximum, name=""):
        super(LessThanFilter, self).__init__(name)
        self.max_level = exclusive_maximum

    def filter(self, record):
        #non-zero return means we log this message
        return 1 if record.levelno < self.max_level else 0


class Logger(SingletonMixin):
    """
    The logger class allows to write in standard output and rotating files the log of executions

    There is 4 logging handler:
      - stream_handler: the logger which will be called to write in std out.
      - stream_handler_err: the logger which will be called to write errors in std out.
      - file_handler: the logger which will be called to write in the .log rotating file.
      - file_handler_err: the logger logger which will write everything in the .err rotating file.

    The file_handler_err should be used for debuging purposes.
    """

    def __init__(self):

        self.__logger = logging.getLogger('wopmars')
        self.formatter_str = '%(asctime)s :: %(levelname)s :: %(name)s :: %(message)s'
        formatter = logging.Formatter(self.formatter_str)
        self.__logger.setLevel(logging.DEBUG)  # set root's level

        verbosity = int(OptionManager.instance()["-v"])

        ################################################################################################################
        #
        # Stream stderr
        #
        ################################################################################################################

        self.stream_handler_stderr = logging.StreamHandler(stream=sys.stderr)
        self.stream_handler_stderr.setFormatter(formatter)
        self.stream_handler_stderr.setLevel(logging.WARNING)
        self.__logger.addHandler(self.stream_handler_stderr)

        ################################################################################################################
        #
        # Stream stdout
        #
        ################################################################################################################

        self.stream_handler_stdout = logging.StreamHandler(stream=sys.stdout)
        self.stream_handler_stdout.setFormatter(formatter)
        if verbosity <= 0:
            self.stream_handler_stdout.setLevel(logging.WARNING)
        if verbosity == 1:
            self.stream_handler_stdout.setLevel(logging.INFO)
        elif verbosity >= 2:
            self.stream_handler_stdout.setLevel(logging.DEBUG)
        self.stream_handler_stdout.addFilter(LessThanFilter(logging.WARNING)) # Does not print in stdout higher/equal than warning
        self.__logger.addHandler(self.stream_handler_stdout)

        ################################################################################################################
        #
        # File stderr
        #
        ################################################################################################################

        if "--log" in OptionManager.instance():

            if not OptionManager.instance()["--log"] is None:

                log_stdout_path = OptionManager.instance()["--log"].rsplit(".", 1)[0]

                # log file in append mode of size 1 Mo and 1 backup
                # handler equivalent to stream_handler in term of logging level but write in .log file
                self.__file_handler_stdout = RotatingFileHandler(log_stdout_path + ".log", 'a', 1000000, 1)
                self.__file_handler_stdout.setFormatter(formatter)
                if verbosity <= 0:
                    self.__file_handler_stdout.setLevel(logging.WARNING)
                if verbosity == 1:
                    self.__file_handler_stdout.setLevel(logging.INFO)
                elif verbosity >= 2:
                    self.__file_handler_stdout.setLevel(logging.DEBUG)
                self.__file_handler_stdout.addFilter(LessThanFilter(logging.WARNING)) # Does not print in stdout higher/equal than warning
                self.__logger.addHandler(self.__file_handler_stdout)

                # err file in append mode of size 1 Mo and 1 backup
                # this handler will write everything in the .err file.
                self.__file_handler_stderr = RotatingFileHandler(log_stdout_path + ".err", 'a', 1000000, 1)
                self.__file_handler_stderr.setFormatter(formatter)
                self.__file_handler_stderr.setLevel(logging.WARNING)
                self.__logger.addHandler(self.__file_handler_stderr)

    def debug(self, msg):
        formatter_stream = logging.Formatter(ColorPrint.blue(self.formatter_str))
        self.stream_handler_stderr.setFormatter(formatter_stream)
        self.stream_handler_stdout.setFormatter(formatter_stream)
        self.__logger.debug(msg)

    def info(self, msg):
        formatter_stream = logging.Formatter(ColorPrint.green(self.formatter_str))
        self.stream_handler_stderr.setFormatter(formatter_stream)
        self.stream_handler_stdout.setFormatter(formatter_stream)
        self.__logger.info(msg)

    def warning(self, msg):
        formatter_stream = logging.Formatter(ColorPrint.yellow(self.formatter_str))
        self.stream_handler_stderr.setFormatter(formatter_stream)
        self.stream_handler_stdout.setFormatter(formatter_stream)
        self.__logger.warning(msg)

    def error(self, msg):
        formatter_stream = logging.Formatter(ColorPrint.red(self.formatter_str))
        self.stream_handler_stderr.setFormatter(formatter_stream)
        self.stream_handler_stdout.setFormatter(formatter_stream)
        self.__logger.error(msg)

    def critical(self, msg):
        formatter_stream = logging.Formatter(ColorPrint.red(self.formatter_str))
        self.stream_handler_stderr.setFormatter(formatter_stream)
        self.stream_handler_stdout.setFormatter(formatter_stream)
        self.__logger.critical(msg)



import os
import pathlib

from sqlalchemy import Column, Integer, String, ForeignKey, DateTime
from sqlalchemy.orm import relationship

from wopmars.Base import Base
from wopmars.SQLManager import SQLManager
from wopmars.models.Option import Option
from wopmars.utils.Logger import Logger
from wopmars.utils.OptionManager import OptionManager
from wopmars.utils.WopMarsException import WopMarsException
from wopmars.utils.various import get_mtime, get_current_time
from wopmars.models.TableModificationTime import TableModificationTime


class ToolWrapper(Base):
    """
    The class ToolWrapper is the superclass of the wrappers which will be designed by the wrapper developers. It is the
    SQLAlchemy Model of the table ``wom_rule`` with the following fields:

    - id: INTEGER - primary_key - auto increment - arbitrary ID
    - rule_name: VARCHAR(255) - the name of the rule
    - tool_python_path: VARCHAR(255) - the is_input of the Toolwrapper
    - execution_id: INTEGER - foreign key to the table ``wom_execution`` - the associated execution
    - started_at: INTEGER - unix mtime_epoch_millis [ms] at wich the tool_python_path started its execution
    - finished_at: INTEGER - unix mtime_epoch_millis [ms] at wich the tool_python_path finished its execution
    - mtime_epoch_millis: FLOAT - the total mtime_epoch_millis [ms] tool_python_path execution
    - status: VARCHAR(255) - the final status of the Toolwrapper. it can be:

       - NOT PLANNED: the tool_python_path execution was not even expected by the user
       - ALREADY EXECUTED: the tool_python_path has been previously executed in an old workflow and does not need to be re-executed
       - EXECUTED: the tool_python_path has been executed
       - ERROR: the tool_python_path has encountered an error during the execution
    """

    __tablename__ = "wom_{}".format(__qualname__)

    id = Column(Integer, primary_key=True, autoincrement=True)
    rule_name = Column(String(255))
    tool_python_path = Column(String(255))
    execution_id = Column(Integer, ForeignKey("wom_Execution.id"))
    started_at = Column(DateTime, nullable=True)
    finished_at = Column(DateTime, nullable=True)
    run_duration_secs = Column(Integer, nullable=True)
    status = Column(String(255), nullable=True, default="NOT_EXECUTED")

    # One rule has many tables
    relation_toolwrapper_to_tableioinfo = relationship("TableInputOutputInformation", back_populates="relation_file_or_tableioinfo_to_toolwrapper", cascade="all, delete, delete-orphan")
    # One rule has many files
    relation_toolwrapper_to_fileioinfo = relationship("FileInputOutputInformation", back_populates="relation_file_or_tableioinfo_to_toolwrapper", cascade="all, delete, delete-orphan")
    # One option is used by many toolwrappers
    relation_toolwrapper_to_option = relationship("Option", back_populates="relation_option_to_toolwrapper", cascade="all, delete, delete-orphan")
    # One rule has one execution
    # execution = relationship("Execution", back_populates="rules")
    relation_toolwrapper_to_execution = relationship("Execution", back_populates="relation_execution_to_toolwrapper")

    # parentrules = relationship etc...
    __mapper_args__ = {
        'polymorphic_on': tool_python_path,
        'polymorphic_identity': 'ToolWrapper'
    }

    NEW = 1
    READY = 2
    NOT_READY = 3

    def __init__(self, rule_name=""):
        """
        The constructor of the tool_python_path, must not be overwritten.

        self.__state is the state given to the Toolwrapper to let the
        :class:`~.wopmars.framework.management.WorflowManager.WorkflowManager` knows if the Toolwrapper is
        able to be executed or not.
        self.session is the session (WopmarsSession) associated with the Toolwrapper and which will be used in the run method.
        self.__state is an integer which says the actual state of the TooLWrapper: ``NEW``, ``READY``, ``NOT_READY``

        :param rule_name: the is_input of the rule
        :type rule_name: str
        """
        super().__init__(rule_name=rule_name)
        self.__state = ToolWrapper.NEW
        self.session = None

    def is_content_respected(self):
        """
        Parsing method:

        This method checks if the parameters dictionary are properly formed, according to the specifications of the
        wrapper developer.

        Call of the methods:

        - :meth:`~.wopmars.framework.database.ToolWrapper.ToolWrapper.is_options_respected`
        - :meth:`~.wopmars.framework.database.ToolWrapper.ToolWrapper.is_input_or_output_respected`
        - :meth:`~.wopmars.framework.database.ToolWrapper.ToolWrapper.is_output_respected`
        """
        # the relation_toolwrapper_to_option have to be checked first because they can alter the behavior of the is_input_or_output_respected and
        # is_output_respected methods
        self.is_options_respected()

        # Check whether input is respected
        self.is_input_or_output_respected(is_input=True)
        # Check whether output is respected
        self.is_input_or_output_respected(is_input=False)

    def is_input_or_output_respected(self, is_input):
        """
        Parsing method:

        Check if the input (is_input=1) or output (is_input=0) file and table variables names associated with the
        tool_wrapper are respected. If not, throws a WopMarsParsingException.

        This method calls the :meth:`~.wopmars.framework.database.ToolWrapper.ToolWrapper.specify_input_file` method
        which have been written by the tool_python_path developer.

        :raise WopMarsException: The input are not respected by the user.
        """

        fileioinfo_name_set = set([fileioinfo.file_key for fileioinfo in self.relation_toolwrapper_to_fileioinfo
                                    if fileioinfo.relation_file_or_tableioinfo_to_typeio.is_input == is_input])

        tableio_tablename_set = set([tableioinfo.table_key for tableioinfo in self.relation_toolwrapper_to_tableioinfo
                                     if tableioinfo.relation_file_or_tableioinfo_to_typeio.is_input == is_input])

        # tableio_tablename_set = set([tableioinfo.model_py_path.split('.')[-1] for tableioinfo in self.relation_toolwrapper_to_tableioinfo
        #                              if tableioinfo.relation_file_or_tableioinfo_to_typeio.is_input == is_input])

        if is_input:

            ############################################################################################################
            #
            # Input files
            #
            ############################################################################################################

            tool_wrapper_specified_files = self.specify_input_file()

            if fileioinfo_name_set != set(tool_wrapper_specified_files):
                msg_err = "The given input file variable names for {} (rule {}) are not correct." \
                          " They should be: {}" \
                          " They are: {}"\
                    .format(self.__class__.__name__, str(self.rule_name), "'\n\t'".join(tool_wrapper_specified_files),
                            "'\n\t'".join(fileioinfo_name_set))
                raise WopMarsException(msg_err)

            ############################################################################################################
            #
            # Input tables
            #
            ############################################################################################################

            tool_wrapper_specified_tables = self.specify_input_table()

            if tableio_tablename_set != set(tool_wrapper_specified_tables):
                msg_err = "The given input file variable names for {} (rule {}) are not correct." \
                          " They should be: {}" \
                          " They are: {}"\
                    .format(self.__class__.__name__, str(self.rule_name), "'\n\t'".join(tool_wrapper_specified_tables),
                            "'\n\t'".join(tableio_tablename_set))
                raise WopMarsException(msg_err)

        else:

            ############################################################################################################
            #
            # Output files
            #
            ############################################################################################################

            tool_wrapper_specified_files = self.specify_output_file()

            if fileioinfo_name_set != set(tool_wrapper_specified_files):
                msg_err = "The given output file variable names for {} (rule {}) are not correct." \
                          " They should be: {}" \
                          " They are: {}"\
                    .format(self.__class__.__name__, str(self.rule_name), "'\n\t'".join(tool_wrapper_specified_files),
                            "'\n\t'".join(fileioinfo_name_set))
                raise WopMarsException(msg_err)

            ############################################################################################################
            #
            # Output tables
            #
            ############################################################################################################

            tool_wrapper_specified_tables = self.specify_output_table()

            if tableio_tablename_set != set(tool_wrapper_specified_tables):
                msg_err = "The given output table variable names for {} (rule {}) are not correct." \
                          " They should be: {}" \
                          " They are: {}"\
                    .format(self.__class__.__name__, str(self.rule_name), "'\n\t'".join(tool_wrapper_specified_tables),
                            "'\n\t'".join(tableio_tablename_set))
                raise WopMarsException(msg_err)

    def is_options_respected(self):
        """
        Parsing method:

        This method check if the params given in the constructor are properly formed for the tool.

        It checks if the params names given by the user exists or not, if the type correspond and if the required
        relation_toolwrapper_to_option are given. If not, throws a WopMarsParsingException.

        This method calls the "specify_params" method of the tool_python_path. This method should return a dictionnary
        associating the is_input of the option with a String containing the types allowed with it. A "|" is used between
        each types allowed for one option.

        Example:

        .. code-block:: python

            {
                'option1': "int",
                'option2': "required|str",
            }

        :raises WopMarsException: If the params names and types are not respected by the user.
        """
        dict_wrapper_opt_carac = self.specify_params()

        # check if the given relation_toolwrapper_to_option are authorized
        if not set([opt.name for opt in self.relation_toolwrapper_to_option]).issubset(dict_wrapper_opt_carac):
            raise WopMarsException("The content of the definition file is not valid.",
                                   "The given option variable for the rule " + str(self.rule_name) + " -> " + self.__class__.__name__ +
                                   " are not correct, they should be in: " +
                                   "\n\t'{0}'".format("'\n\t'".join(dict_wrapper_opt_carac)) +
                                   "\n" + "They are:" +
                                   "\n\t'{0}'".format("'\n\t'".join([opt.name for opt in self.relation_toolwrapper_to_option]))
                                   )

        # check if the types correspond
        for opt in self.relation_toolwrapper_to_option:
            opt.correspond(dict_wrapper_opt_carac[opt.name])

        # check if the required relation_toolwrapper_to_option are given
        for opt in dict_wrapper_opt_carac:
            if "required" in str(dict_wrapper_opt_carac[opt]).lower() and opt not in [opt2.name for opt2 in self.relation_toolwrapper_to_option]:
                raise WopMarsException("The content of the definition file is not valid.",
                                       "The option '" + opt + "' has not been provided but it is required.")

    def follows(self, other):
        """
        Parsing method:

        Check whether the "self" follows directly "other" in the execution DAG.

        Check whether "other" has one output value in "self" possible input values.
        The output value are given from the Relationnal mapping between toolwrappers and related objects:
        * TableInputOutputInformation
        * FileInputOutputInformation

        :param other: ToolWrapper that is possibly a predecessor of "self"
        :return: bool True if "self" follows "other"
        """
        for rule_f_path in [f.path for f in self.relation_toolwrapper_to_fileioinfo if f.relation_file_or_tableioinfo_to_typeio.is_input == 1]:
            for rule_f2_path in [f.path for f in other.relation_toolwrapper_to_fileioinfo if f.relation_file_or_tableioinfo_to_typeio.is_input == 0]:
                if rule_f_path == rule_f2_path:
                    return True

        for rule_t_name in [t.model_py_path for t in self.relation_toolwrapper_to_tableioinfo if t.relation_file_or_tableioinfo_to_typeio.is_input == 1]:
            for rule_t2_name in [t.model_py_path for t in other.relation_toolwrapper_to_tableioinfo if t.relation_file_or_tableioinfo_to_typeio.is_input == 0]:
                if rule_t_name == rule_t2_name:
                    return True

        return False

    ### Workflow Manager methods

    def get_input_files_not_ready(self):
        """
        Check if inputs are ready

        :return: bool - True if inputs are ready.
        """
        input_files_not_ready = []
        input_files = [f for f in self.relation_toolwrapper_to_fileioinfo if f.relation_file_or_tableioinfo_to_typeio.is_input == 1]
        for i in input_files:
            if not i.is_ready():
                input_files_not_ready.append(i)
        return input_files_not_ready

    def are_inputs_ready(self):
        """
        Check if inputs are ready

        :return: bool - True if inputs are ready.
        """
        input_files = [f for f in self.relation_toolwrapper_to_fileioinfo if f.relation_file_or_tableioinfo_to_typeio.is_input == 1]
        Logger.instance().debug("Inputs files of " + str(self.__class__.__name__) + ": " + str([i.file_key for i in input_files]))
        for i in input_files:
            if not i.is_ready():
                Logger.instance().debug("Input: " + str(i.file_key) + " is not ready.")
                self.__state = ToolWrapper.NOT_READY
                return False
            Logger.instance().debug("Input: " + str(i.file_key) + " is ready.")

        input_tables = [t for t in self.relation_toolwrapper_to_tableioinfo if t.relation_file_or_tableioinfo_to_typeio.is_input == 1]
        Logger.instance().debug("Inputs tables of " + str(self.__class__.__name__) + ": " + str([i.table_key for i in input_tables]))
        for i in input_tables:
            if not i.is_ready():
                Logger.instance().debug("Input: " + str(i.table_key) + " is not ready.")
                self.__state = ToolWrapper.NOT_READY
                return False
            Logger.instance().debug("Input: " + str(i.table_key) + " is ready.")

        self.__state = ToolWrapper.READY
        return True

    def set_args_time_and_size(self, is_input, dry=False):
        """
        WorkflowManager method:

        The mtime_epoch_millis and the size of the files are set according to the actual mtime_epoch_millis of last modification and size of the system files

        The mtime_epoch_millis of the tables are set according to the mtime_epoch_millis of last modification notified in the modification_table table
        If the is_input of InputOutput is "output" and the execution is "not dry", the mtime_epoch_millis in modification_table is set to the
        current mtime_epoch_millis.mtime_epoch_millis().

        # totodo LucG modify it to take commits into account instead of the status of 'output' of a table

        :param is_input: "input" or "output"
        :is_input is_input: bool
        :param dry: Say if the execution has been simulated.
        :is_input dry: bool
        """
        session = SQLManager.instance().get_session()
        for f in [f for f in self.relation_toolwrapper_to_fileioinfo if f.relation_file_or_tableioinfo_to_typeio.is_input == is_input]:
            try:
                mtime_epoch_millis, mtime_human = get_mtime(f.path)
                f.mtime_human = mtime_human
                f.mtime_epoch_millis = mtime_epoch_millis
                size = os.path.getsize(f.path)
                f.size = size
                size = os.path.getsize(f.path)
            except FileNotFoundError as FE:
                # totodo LucG ask lionel sans ce rollback, ca bug, pourquoi? la session est vide... comme si la query etait bloquante
                if not OptionManager.instance()["--dry-run"]:
                    session.rollback()
                    raise WopMarsException("Error during the execution of the workflow",
                                           "The " + is_input + " file " + str(f.path) + " of rule " + str(self.rule_name) +
                                           " doesn't exist")
                else:
                    # in dry-run mode, input/output files might not exist
                    mtime_epoch_millis = None
                    size = None
            f.used_at = mtime_epoch_millis
            f.size = size
            session.add(f)
            if is_input == 1:
                Logger.instance().debug("Input file " + str(f) + " used.")
            elif is_input == 0 and dry:
                Logger.instance().debug("Output file " + str(f) + " has been loaded from previous execution.")
            elif is_input == 0 and not dry:
                Logger.instance().debug("Output file " + str(f) + " has been created.")
        # this commit is due to a bug that i couldn't figure out: the session empty itself between the two loops...
        # this is not good at all since it may lead to inconsistence in the database
        session.commit()

        for t in [tableioinfo for tableioinfo in self.relation_toolwrapper_to_tableioinfo
                  if tableioinfo.relation_file_or_tableioinfo_to_typeio.is_input == is_input]:
            t.mtime_human = t.relation_tableioinfo_to_tablemodiftime.mtime_human
            t.mtime_epoch_millis = t.relation_tableioinfo_to_tablemodiftime.mtime_epoch_millis
            # t.used_at = t.relation_tableioinfo_to_tablemodiftime.mtime_epoch_millis
            session.add(t)
        session.commit()

    def same_input_than(self, other):
        """
        Never used.

        Check if the other ToolWrapper have the same inputs than self.

        The input are say "the same" if:
            - The table have the same is_input and the same last modification mtime_epoch_millis
            - The file have the same is_input, the same last modification mtime_epoch_millis and the same size

        :param other: an other Toolwrapper which maybe as the same inputs
        :type other: :class:`~.wopmars.framework.database.relation_toolwrapper_to_tableioinfo.ToolWrapper.ToolWrapper`

        :return: bool
        """

        #############################################################
        #
        # Check file input
        #
        #############################################################

        for f in [f for f in self.relation_toolwrapper_to_fileioinfo if f.relation_file_or_tableioinfo_to_typeio.is_input == 1]:
            is_same = False
            for f2 in [f2 for f2 in other.relation_toolwrapper_to_fileioinfo if f2.relation_file_or_tableioinfo_to_typeio.is_input == 1]:
                # two files are the same if they have the same is_input, path, size and modification mtime_epoch_millis
                if (f.file_key == f2.file_key and f.path == f2.path and
                        f.mtime_epoch_millis == f2.mtime_epoch_millis and f.size == f2.size):
                    is_same = True
                    break
            if not is_same:
                return False

        #############################################################
        #
        # Check table input
        #
        #############################################################

        for t in [t for t in self.relation_toolwrapper_to_tableioinfo if t.relation_file_or_tableioinfo_to_typeio.is_input == 1]:
            is_same = False
            for t2 in [t2 for t2 in other.relation_toolwrapper_to_tableioinfo if t2.relation_file_or_tableioinfo_to_typeio.is_input == 1]:
                # two tables are the same if they have the same model/table_key/modification mtime_epoch_millis
                if (t.model_py_path == t2.model_py_path and t.table_key == t2.table_key and
                       t.mtime_epoch_millis == t2.mtime_epoch_millis):
                    is_same = True
                    break
            if not is_same:
                return False

        return True

    def is_output_more_recent_than_input(self):
        """
        Check for files and tables if the outputs are more recent than inputs.

        In a conventionnal use of WoPMaRS, the output are supposed to be younger than the inputs. If they are not,
        we can consider that the input has changed since the last execution and the output has to be re-written.

        :return: Bool: True if the output is actually more recent than input
        """

        newest_input_file = [get_mtime(input_fileioinfo.path)[0]
                             for input_fileioinfo in self.relation_toolwrapper_to_fileioinfo
                                  if input_fileioinfo.relation_file_or_tableioinfo_to_typeio.is_input == 1]
        newest_input_table = [input_tableioinfo.relation_tableioinfo_to_tablemodiftime.mtime_epoch_millis
                                   for input_tableioinfo in self.relation_toolwrapper_to_tableioinfo
                                   if input_tableioinfo.relation_file_or_tableioinfo_to_typeio.is_input == 1]
        # newest is supposed to have largest epoch time
        newest_input = max(newest_input_file + newest_input_table)

        oldest_output_file = [get_mtime(output_fileioinfo.path)[0]
                              for output_fileioinfo in self.relation_toolwrapper_to_fileioinfo
                              if output_fileioinfo.relation_file_or_tableioinfo_to_typeio.is_input == 0]
        oldest_output_table = [output_tableioinfo.relation_tableioinfo_to_tablemodiftime.mtime_epoch_millis
                               for output_tableioinfo in self.relation_toolwrapper_to_tableioinfo
                               if output_tableioinfo.relation_file_or_tableioinfo_to_typeio.is_input == 0]
        # newest is supposed to have smallest epoch time
        oldest_output = min(oldest_output_file + oldest_output_table)

        return newest_input < oldest_output

    def same_output_than(self, other):
        """
        Never used.

        Check if the output of self is the same than other.

        Checks only if the file names, table names and models are the same.

        :return: bool
        """
        for t in [t for t in self.relation_toolwrapper_to_tableioinfo if t.relation_file_or_tableioinfo_to_typeio.is_input == 0]:
            is_same = False
            for t2 in [t2 for t2 in other.relation_toolwrapper_to_tableioinfo if t2.relation_file_or_tableioinfo_to_typeio.is_input == 0]:
                if t.model_py_path == t2.model_py_path and t.table_key == t2.table_key:
                    is_same = True
                    break
            if not is_same:
                return False

        for f in [f for f in self.relation_toolwrapper_to_fileioinfo if f.relation_file_or_tableioinfo_to_typeio.is_input == 1]:
            is_same = False
            for f2 in [f2 for f2 in other.relation_typeio_to_fileioinfo if f2.relation_file_or_tableioinfo_to_typeio.is_input == 1]:
                if (f.file_key == f2.file_key and
                        f.path == f2.path):
                    is_same = True
                    break
            if not is_same:
                return False
        return True

    def output_file_exists(self):
        """
        Check if the file exists in the system.

        :return: Bool: True if outputs exist.
        """
        for output_file in [this_file for this_file in self.relation_toolwrapper_to_fileioinfo
                            if this_file.relation_file_or_tableioinfo_to_typeio.is_input == 0]:
            if not os.path.exists(output_file.path):
                return False
        return True

    def output_table_exists(self):
        """
        Check if the table is not empty.

        :return: Bool: True if outputs exist.
        """

        for output_table in [table for table in self.relation_toolwrapper_to_tableioinfo
                             if table.relation_file_or_tableioinfo_to_typeio.is_input == 0]:
            if not SQLManager.instance().get_session().query(output_table.get_table()).count():
                return False
        return True

    def get_state(self):
        return self.__state

    def set_execution_infos(self, start=None, stop=None, status=None):
        """
        Generic method to set the informations relatives to the execution of the ToolWrapper.

        :param start: The mtime_epoch_millis of start of the Toolwrapper
        :param stop: The mtime_epoch_millis of end of the Toolwrapper
        :param status: The status of the Toolwrapper
        """
        if start is not None:
            self.started_at = start
        if stop is not None:
            self.finished_at = stop
        if self.started_at is not None and self.finished_at is not None:
            self.run_duration_secs = (self.finished_at - self.started_at).total_seconds()
        if status is not None:
            self.status = status

    def __eq__(self, other):
        """
        Two tool_wrappers are equal for these conditions:

        - Both belong to the same class
        - Both have the same input and output fields

        We check if the files, tables and relation_toolwrapper_to_option are the same.
        :param other: ToolWrapper
        :type other: ToolWrapper
        :return: Bool: True if the ToolWrappers are equals.
        """
        return (isinstance(other, self.__class__) and
                self.same_files(other, is_input=True) and
                self.same_tables(other, is_input=True) and
                self.same_files(other, is_input=False) and
                self.same_tables(other, is_input=False) and
                self.same_options(other))

    def same_files(self, other, is_input):
        """
        Check if the files of a ToolWrapper are the same than the files of the other for a given type (input or output).

        :param other: ToolWrapper with which you need to compare
        :type other: ToolWrapper
        :param is_input: The is_input of the type of file (input or output)
        :type is_input: str
        :return: Bool: True if the files are the same
        """
        for f in [rf for rf in self.relation_toolwrapper_to_fileioinfo if rf.relation_file_or_tableioinfo_to_typeio.is_input == is_input]:
            is_in = bool([rf for rf in other.relation_toolwrapper_to_fileioinfo if (
                    os.path.abspath(f.path) == os.path.abspath(rf.path) and # same absolute path
                    f.file_key == rf.file_key and  # same file field name
                    rf.relation_file_or_tableioinfo_to_typeio.is_input == is_input  # files are same input
            )])
            if not is_in:
                return False
        return True

    def same_tables(self, other, is_input):
        """
        Check if the tables of a ToolWrapper are the same than the tables of the other for a given type (input or output).

        :param other: ToolWrapper with which you need to compare
        :type other: ToolWrapper
        :param is_input: The is_input of the type of table (input or output)
        :type is_input: str
        :return: Bool: True if the tables are the same
        """
        for t in [t for t in self.relation_toolwrapper_to_tableioinfo if t.relation_file_or_tableioinfo_to_typeio.is_input == is_input]:
            is_in = bool([t for t in other.relation_toolwrapper_to_tableioinfo if (t.model_py_path == t.model_py_path and
                                                                                   t.relation_file_or_tableioinfo_to_typeio.is_input == is_input and
                                                                                   t.table_key == t.table_key)])
            if not is_in:
                return False
        return True

    def same_options(self, other):
        """
        Check if the relation_toolwrapper_to_option of a ToolWrapper are the same the relation_toolwrapper_to_option of the other.

        :param other: ToolWrapper with which you need to compare.
        :type other: ToolWrapper
        :return: Bool: True if the relation_toolwrapper_to_option are the same.
        """
        for opt in self.relation_toolwrapper_to_option:
            is_in = bool([o for o in other.relation_toolwrapper_to_option if (o.name == opt.name and
                                                       o.value == opt.value)])

            if not is_in:
                return False
        return True

    def __hash__(self):
        """
        Redefining the hash method allows ToolWrapper objects to be indexed in sets and dict.

        Needed to use ToolWrapper as nodes of the DiGraph.

        :return: int
        """
        return id(self)

    def __repr__(self):
        """
        Return the string representing the tool_python_path in the DAG.

        :return: String representing the tool_python_path
        """
        s = "\""
        s += "ToolWrapper " + self.rule_name
        s += "\\n"
        s += "tool: " + self.__class__.__name__
        s += "\\n"
        for input_f in [f for f in self.relation_toolwrapper_to_fileioinfo if f.relation_file_or_tableioinfo_to_typeio.is_input == 1]:
            s += "\\n\t\t" + input_f.file_key + ": " + str(input_f.path)
        for input_t in [t for t in self.relation_toolwrapper_to_tableioinfo if t.relation_file_or_tableioinfo_to_typeio.is_input == 1]:
            s += "\\n\t\tinput_table: " + input_t.table_key
        s += "\\n"
        for output_f in [f for f in self.relation_toolwrapper_to_fileioinfo if f.relation_file_or_tableioinfo_to_typeio.is_input == 0]:
            s += "\\n\t\t" + output_f.file_key + ": " + str(output_f.path)
        for output_t in [t for t in self.relation_toolwrapper_to_tableioinfo if t.relation_file_or_tableioinfo_to_typeio.is_input == 0]:
            s += "\\n\t\toutput_table: " + output_t.table_key
        s += "\""
        return s

    def dot_label(self):
        """Label for the dot dag"""
        inputs_list_str = [str(i).replace(":", "") for i in self.relation_toolwrapper_to_fileioinfo + self.relation_toolwrapper_to_tableioinfo if i.relation_file_or_tableioinfo_to_typeio.is_input == 1]
        outputs_list_str = [str(o).replace(":", "") for o in self.relation_toolwrapper_to_fileioinfo + self.relation_toolwrapper_to_tableioinfo if o.relation_file_or_tableioinfo_to_typeio.is_input == 0]
        params_list_str = [str(p).replace(":","") for p in self.relation_toolwrapper_to_option]
        s = ""
        s += "ToolWrapper " + self.rule_name + "\n"
        s += "ToolWrapper " + self.__class__.__name__ + "\n"
        s += "Inputs\n" + "\n\t".join(inputs_list_str) + "\n"
        s += "Outputs\n" + "\n".join(outputs_list_str) + "\n"
        s += "Parameters\n" + "\n".join(params_list_str) + "\n"
        return(s)

    def __str__(self):
        inputs_list_str = [str(i) for i in self.relation_toolwrapper_to_fileioinfo +
                           self.relation_toolwrapper_to_tableioinfo if i.relation_file_or_tableioinfo_to_typeio.is_input == 1]
        outputs_list_str = [str(o) for o in self.relation_toolwrapper_to_fileioinfo +
                            self.relation_toolwrapper_to_tableioinfo if o.relation_file_or_tableioinfo_to_typeio.is_input == 0]
        params_list_str = [str(p) for p in self.relation_toolwrapper_to_option]
        s = ""
        s += "ToolWrapper " + str(self.rule_name) + ":" + "\n"
        s += "\ttool: " + str(self.tool_python_path) + "\n"
        if len(inputs_list_str) > 0:
            s += "\tinput:" + "\n"
            s += "\t\t" + "\n\t\t".join(inputs_list_str)
            s += "\n"
        if len(outputs_list_str) > 0:
            s += "\toutput:" + "\n"
            s += "\t\t" + "\n\t\t".join(outputs_list_str)
            s += "\n"
        if len(params_list_str) > 0:
            s += "\tparams:" + "\n"
            s += "\t\t" + "\n\t\t".join(params_list_str)
        return s

    def touch(self):
        """Updates modification time of output table or file"""

        ################################################################################################################
        #
        # Touch output table
        #
        ################################################################################################################

        for output_table in [table for table in self.relation_toolwrapper_to_tableioinfo
                             if table.relation_file_or_tableioinfo_to_typeio.is_input == 0]:
            out_table_inst = SQLManager.instance().get_session().query(TableModificationTime).filter_by(table_name=output_table.table_key).first()
            timestamp_epoch_millis, timestamp_human = get_current_time()
            out_table_inst.mtime_epoch_millis = timestamp_epoch_millis
            out_table_inst.mtime_human = timestamp_human
            SQLManager.instance().get_session().commit()

        ################################################################################################################
        #
        # Touch output file
        #
        ################################################################################################################

        for output_file in [this_file for this_file in self.relation_toolwrapper_to_fileioinfo
                            if this_file.relation_file_or_tableioinfo_to_typeio.is_input == 0]:
            if not (OptionManager.instance()["--directory"] is None):
                output_file_path = os.path.join(OptionManager.instance()["--directory"], output_file.path)
            else:
                output_file_path = output_file.path
            if os.path.exists(output_file_path):
                pathlib.Path(output_file_path).touch(exist_ok=True)

    ################################################################################################################
    #
    # Methods that the developer should implement
    #
    ################################################################################################################

    def specify_input_file(self):
        """
        Should be implemented by the tool_python_path developer.

        This method return a List of string containing the input file variable names as String.
        :return: [String]
        """
        return []

    def specify_input_table(self):
        """
        Should be implemented by the tool_python_path developer.

        This method return a List of string containing the input table names names as String.
        :return: [String]
        """
        return []

    def specify_output_file(self):
        """
        Should be implemented by the tool_python_path developer.

        This method return a List of string containing the output file variable names as String.
        :return: [String]
        """
        return []

    def specify_output_table(self):
        """
        Should be implemented by the tool_python_path developper.

        This method return a List of string containing the output table names as String.
        :return: [String]
        """
        return []

    def specify_params(self):
        """
        Should be implemented by the tool_python_path developper.

        This method return a dict of string associated with string. Keys are the is_input of the relation_toolwrapper_to_option and values, their types.

        :return: {String: String}
        """
        return {}

    def run(self):
        """
        Should be implemented by the tool_python_path developer.

        The core function of the ToolWrapper is this method. It wraps the actual execution of the tool underlying the ToolWrapper.

        :raises NotImplementedError: If it doesn't have been implemented by the ToolWrapper Developer.
        """
        raise NotImplementedError("The method run of the ToolWrapper " + str(self.tool_python_path) + " should be implemented")

    ### Methods availables for the tool developer

    def input_file(self, key):
        """
        Return the path of the specified input file.

        :param key: String the is_input of the variable containing the path
        :return:
        """
        try:
            return [f.path for f in self.relation_toolwrapper_to_fileioinfo if f.file_key == key and f.relation_file_or_tableioinfo_to_typeio.is_input == 1][0]
        except IndexError:
            raise WopMarsException("Error during the execution of the ToolWrapper " + str(self.tool_python_path) +
                                   " (rule " + self.rule_name + ").",
                                   "The input file " + str(key) + " has not been specified.")

    def input_table(self, key):
        """
        Return the input table object of the given is_input.

        :param key: String: the is_input of the Table object.
        :return:
        """
        try:
            return [t for t in self.relation_toolwrapper_to_tableioinfo if t.table_key == key and t.relation_file_or_tableioinfo_to_typeio.is_input == 1][0].get_table()
        except IndexError:
            raise WopMarsException("Error during the execution of the ToolWrapper " + str(self.tool_python_path) +
                                   " (rule " + self.rule_name + ").",
                                   "The input table " + str(key) + " has not been specified.")

    def output_file(self, key):
        """
        Return the path of the specified output file.

        :param key: String the is_input of the variable containing the path
        :return:
        """
        try:
            return [f.path for f in self.relation_toolwrapper_to_fileioinfo if f.file_key == key and f.relation_file_or_tableioinfo_to_typeio.is_input == 0][0]
        except IndexError:
            raise WopMarsException("Error during the execution of the ToolWrapper " + str(self.tool_python_path) +
                                   " (rule " + self.rule_name + ").",
                                   "The output file " + str(key) + " has not been specified.")

    def output_table(self, key):
        """
        Return the output table object of the given is_input.

        :param key: String: the is_input of the Table object.
        :return:
        """
        try:
            return [t for t in self.relation_toolwrapper_to_tableioinfo if t.table_key == key and t.relation_file_or_tableioinfo_to_typeio.is_input == 0][0].get_table()
        except IndexError:
            raise WopMarsException("Error during the execution of the ToolWrapper " + str(self.tool_python_path) +
                                   " (rule " + self.rule_name + ").",
                                   "The output table " + str(key) + " has not been specified.")

    def option(self, key):
        """
        Return the value associated with the key option.

        If no value is associated with key, return None.

        :param key: The is_input of the option
        :type key: str
        :return:
        """
        try:
            value = [o.value for o in self.relation_toolwrapper_to_option if o.name == key][0]
            list_splitted_carac = self.specify_params()[key].split("|")
            for s_type in list_splitted_carac:
                s_formated_type = s_type.strip().lower()
                # check if the carac is a castable type
                if s_formated_type in Option.static_option_castable:
                    value = eval(s_formated_type)(value)
                    break
            return value
        except IndexError as e:
            # antipattern, this is bad, but I deleted the warning because if the ToolWrapper Developer put his call to
            # option in a loop, there will be too mutch output
            pass
            return None

    def session(self):
        return self.session



from unittest import TestCase
from wopmars.utils.PathManager import PathManager
import os
import tempfile
import unittest


class TestPathFinder(TestCase):

    def test_check_valid_path(self):
        self.assertIsNone(PathManager.check_valid_path(None))
        self.assertRaises(FileNotFoundError, PathManager.check_valid_path, ("/coucou/toto"))

        tempdir = os.path.join(tempfile.gettempdir(), 'test_bak')
        self.assertEqual(PathManager.check_valid_path(tempdir), tempdir)


if __name__ == '__main__':
    unittest.main()



# -*- coding: UTF-8 -*-

# __author__ = "Aitor Gonzalez, Luc Giffon, Lionel Spinelli"
# __copyright__ = "Copyright: since 2017, Aitor Gonzalez, Luc Giffon, Lionel Spinelli"
# __email__ = "aitor.gonzalez@univ-amu.fr"
# __license__ = "MIT"
import codecs
from codecs import open
from os import path
import sys
import os
import configparser

try:
    from setuptools import setup, find_packages
except ImportError:
    print("Please install setuptools before installing snakemake.", file=sys.stderr)
    exit(1)

if sys.version_info < (3, 6):
    print("Python version >= 3.6 required.")

here = os.path.abspath(os.path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as fin:
    long_description = fin.read()

CLASSIFIERS = """\
Development Status :: 4 - Beta
Environment :: Console
Intended Audience :: Developers
Intended Audience :: Science/Research
License :: OSI Approved :: MIT License
Programming Language :: Python
Programming Language :: Python :: 3
Programming Language :: Python :: 3.6
Programming Language :: Python :: 3.7
Programming Language :: Python :: 3.8
Programming Language :: Python :: 3 :: Only
Topic :: Database :: Front-Ends
Topic :: Scientific/Engineering :: Bio-Informatics
Topic :: Scientific/Engineering :: Information Analysis
Topic :: Software Development
Operating System :: POSIX :: Linux
Operating System :: Microsoft :: Windows :: Windows 10
"""

def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()

def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")

# Create list of package data files
def data_files_to_list(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths

data_file_list = data_files_to_list('wopmars/data')

config = configparser.RawConfigParser()
config.read(os.path.join('.', 'setup.cfg'))
author = config['metadata']['author']
email = config['metadata']['email']
license = config['metadata']['license']

setup(name='wopmars',
      version=get_version("wopmars/__init__.py"),
      description='Workflow Python Manager for Reproducible Science',
      long_description=long_description,
      url='https://github.com/aitgon/wopmars',
      author=author,
      author_email=email,
      license=license,
      classifiers=[_f for _f in CLASSIFIERS.split('\n') if _f],
      download_url='https://github.com/aitgon/wopmars/archive/%s.tar.gz'%(get_version("wopmars/__init__.py")),
      keywords='workflow manager python object-oriented reproducible science database framework',
      extras_require={'pygraphviz': ['pygraphviz',]},
      data_files=[],
      packages=find_packages(exclude=['log', 'doc', '*.test_bak.*', 'soutenance', 'rapport']),
      package_dir={'wopmars': 'wopmars'},
      package_data={'wopmars': data_file_list},
      include_package_data=True,
      install_requires=["pyyaml", "sqlalchemy", "docopt", "networkx", "schema", "termcolor"],
      entry_points={'console_scripts': ['wopmars=wopmars:run']},
      )




"""
Module containing the SQLManager class.
"""
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy import create_engine, select
from sqlalchemy.schema import sort_tables
from sqlalchemy import event

from wopmars.Base import Base
from wopmars.WopmarsSession import WopmarsSession
from wopmars.utils.Logger import Logger
from wopmars.utils.OptionManager import OptionManager
from wopmars.utils.RWLock import RWLock
from wopmars.utils.Singleton import SingletonMixin
from wopmars.utils.WopMarsException import WopMarsException


class SQLManager(SingletonMixin):
    """
    The class SQLManager is responsible of all operations performed on the database.

    The SQLManager is a synchronized Singleton which may be accessible from any thread, at any mtime_epoch_millis. This is allowed
    by the SingletonMixin superclass and a iterate_wopfile_yml_dic_and_insert_rules_in_db/write lock.

    Provides a WopmarsSession through the method "get_session". The session is used to perform calls to the database.
    Provides the engine if necessary through "get_engine".
    Allows commiting and rollbacking operations of a given session through methods "commit" and "rollback".
    Allows querying operations from a given query: all, one, first, count, one_or_none, scalar. Through "result_factory".
    Allows executing statements through the "execute" method.
    Allows droping and creating tables from table names ("create" and "drop"), list of table models ("drop_table_list")
    and all ("create_all" and "drop_all").
    """
    wopmars_history_tables = [
        "wom_TypeInputOrOutput",
        "wom_TableModificationTime",
        "wom_Execution",
        "wom_ToolWrapper",
        "wom_TableInputOutputInformation",
        "wom_FileInputOutputInformation",
        "wom_Option"
    ]

    def __init__(self):
        """
        The constructor of the SQLManager is supposed to be called once in the whole execution, thanks to the
        SingletonMixin inheritance.

        The constructor create the engine at the localization provided by the user. The option "connect_args" is set to
        False.

        The "PRAGMA foreign_keys=ON" statement is executed here and allows to enforce foreign_keys constraints.

        The Session attribute is a object of class "Type" and allows to make a scoped session bound to the engine on demand.
        The autocommit is set to False and the autoflush, to True.

        The lock is initialized here thanks to the RW lock class and will be used to overide the behaviour of sqlite
        to assess the access to the databse without using the queue of SQLite, bound at 4 sec wait before error.
        """
        s_database_url = OptionManager.instance()["--database"]

        self.d_database_config = {'db_connection': None, 'db_database': None, 'db_url': s_database_url}
        self.d_database_config['db_connection'] = s_database_url.split("://")[0]
        if self.d_database_config['db_connection'] == "sqlite":
            self.d_database_config['db_database'] = s_database_url.replace("sqlite:///", "")
        # echo=False mute the log of database
        # connect_args have been necessary because of the accession of the same objects in different Threads.
        self.engine = None
        if self.d_database_config['db_connection']=="sqlite":
            self.engine = create_engine(self.d_database_config['db_url'], echo=False, connect_args={'check_same_thread': False})
        else:
            self.engine = create_engine(self.d_database_config['db_url'], echo=False)

        def _fk_pragma_on_connect(dbapi_con, con_record):
            if self.d_database_config['db_connection'] == "sqlite":
                dbapi_con.execute('pragma foreign_keys=ON')

        event.listen(self.engine, 'connect', _fk_pragma_on_connect)
        ###

        # Luc: I don't know why I have used the autoflush=True
        self.__Session = scoped_session(sessionmaker(bind=self.engine, autoflush=True, autocommit=False))
        # The lock
        self.__lock = RWLock()

    def clean_up_unexecuted_tool_wrappers(self):

        from wopmars.models.Option import Option
        from wopmars.models.TableInputOutputInformation import TableInputOutputInformation
        from wopmars.models.FileInputOutputInformation import FileInputOutputInformation
        from wopmars.models.ToolWrapper import ToolWrapper
        # from wopmars.models.Execution import Execution

        if self.engine.has_table(ToolWrapper.__tablename__):

            toolwrapper_result_tuple =  self.engine.execute(select([ToolWrapper.__table__.c.id]).where(ToolWrapper.status == 'NOT_EXECUTED')).fetchall()
            for toolwrapper_id_row in toolwrapper_result_tuple:
                self.engine.execute(Option.__table__.delete().where(Option.toolwrapper_id == toolwrapper_id_row[0]))
            for toolwrapper_id_row in toolwrapper_result_tuple:
                self.engine.execute(TableInputOutputInformation.__table__.delete().where(TableInputOutputInformation.toolwrapper_id == toolwrapper_id_row[0]))
            for toolwrapper_id_row in toolwrapper_result_tuple:
                self.engine.execute(FileInputOutputInformation.__table__.delete().where(FileInputOutputInformation.toolwrapper_id == toolwrapper_id_row[0]))
            for toolwrapper_id_row in toolwrapper_result_tuple:
                self.engine.execute(ToolWrapper.__table__.delete().where(ToolWrapper.id == toolwrapper_id_row[0]))

            toolwrapper_result_tuple = self.engine.execute(select([ToolWrapper.__table__.c.id]).where(ToolWrapper.status == 'ALREADY_EXECUTED')).fetchall()
            for toolwrapper_id_row in toolwrapper_result_tuple:
                self.engine.execute(Option.__table__.delete().where(Option.toolwrapper_id == toolwrapper_id_row[0]))
            for toolwrapper_id_row in toolwrapper_result_tuple:
                self.engine.execute(TableInputOutputInformation.__table__.delete().where(TableInputOutputInformation.toolwrapper_id == toolwrapper_id_row[0]))
            for toolwrapper_id_row in toolwrapper_result_tuple:
                self.engine.execute(FileInputOutputInformation.__table__.delete().where(FileInputOutputInformation.toolwrapper_id == toolwrapper_id_row[0]))
            for toolwrapper_id_row in toolwrapper_result_tuple:
                self.engine.execute(ToolWrapper.__table__.delete().where(ToolWrapper.id == toolwrapper_id_row[0]))

        # if self.engine.has_table(Execution.__tablename__):
        #
        #         self.engine.execute(Execution.__table__.delete().where(Execution.status == None))

    def clear_wopmars_history(self):
        """
        Empty all wom tables except wom_type_input_or_output

        :param list_str_table: [String] the rule_name of the models.
        """
        from wopmars.models.Option import Option
        from wopmars.models.TableInputOutputInformation import TableInputOutputInformation
        from wopmars.models.FileInputOutputInformation import FileInputOutputInformation
        from wopmars.models.TableModificationTime import TableModificationTime
        from wopmars.models.ToolWrapper import ToolWrapper
        from wopmars.models.Execution import Execution

        if self.engine.has_table(Option.__tablename__):
            self.engine.execute(Option.__table__.delete())
        if self.engine.has_table(TableInputOutputInformation.__tablename__):
            self.engine.execute(TableInputOutputInformation.__table__.delete())
        if self.engine.has_table(FileInputOutputInformation.__tablename__):
            self.engine.execute(FileInputOutputInformation.__table__.delete())
        if self.engine.has_table(TableModificationTime.__tablename__):
            self.engine.execute(TableModificationTime.__table__.delete())
        if self.engine.has_table(ToolWrapper.__tablename__):
            self.engine.execute(ToolWrapper.__table__.delete())
        if self.engine.has_table(Execution.__tablename__):
            self.engine.execute(Execution.__table__.delete())

    def get_session(self):
        """
        Return a WopmarsSession associated with this SQLmanager.
        Aitor: I do not know why by this method is necessary rather than self.session direct access
        for test_SQLManager to work

        :return: WopmarsSession
        """
        session = WopmarsSession(self.__Session(), self)
        return session

    def commit(self, session):
        """
        Commit the current session.

        The SQLManager wrap the sqlite queue for commiting operations on database in order to do not trigger the error due
        to the mtime_epoch_millis-out operations.

        This is done thanks to a read_write_lock: sqlmanager is a synchronized singleton with synchronized methods.

        :param session: SQLAlchemy session object
        """
        try:
            Logger.instance().debug(str(session) + " want the write lock on SQLManager.")
            self.__lock.acquire_write()
            Logger.instance().debug(str(session) + " has taken the write lock on SQLManager.")
            session.commit()
            Logger.instance().debug(str(session) + " has been commited.")
        finally:
            # Always release the lock
            self.__lock.release()
            Logger.instance().debug(str(session) + " has released the write lock on SQLManager.")

    def result_factory(self, query, method):
        """
        Return the result of the query, using the demanded method.

        The result_factory wrap the methods for querying database:
          - all
          - first
          - one
          - one_or_none
          - scalar
          - count

        This is necessary to use the ReadLock of WopMars instead of the one from SQLite.

        :param query: The query object, ready to be performed.
        :param method: String signifying the method which have to be used for querying database
        :return: The result of the query.
        """
        result = None
        try:
            Logger.instance().debug("Executing query on session " + str(query.session) + ": \n" + str(query) + ";")
            Logger.instance().debug("WopmarsQuery " + str(query.session) + " want the iterate_wopfile_yml_dic_and_insert_rules_in_db-lock on SQLManager")
            self.__lock.acquire_read()
            Logger.instance().debug("\"" + str(query.session) + "\" has taken the iterate_wopfile_yml_dic_and_insert_rules_in_db lock on SQLManager.")
            # switch case according to the demanded method.
            # in each condition, we call the superclass associated method: superclass is Query from sqlalchemy
            if method == "all":
                result = super(query.__class__, query).all()
            elif method == "one":
                result = super(query.__class__, query).one()
            elif method == "first":
                result = super(query.__class__, query).first()
            elif method == "count":
                result = super(query.__class__, query).count()
            elif method == "one_or_none":
                result = super(query.__class__, query).one_or_none()
            elif method == "scalar":
                result = super(query.__class__, query).scalar()
            else:
                raise WopMarsException("Error while querying the database.",
                                       "Demanded operation doesn't exist: " + str(method))
        finally:
            # Always release the lock
            self.__lock.release()
            Logger.instance().debug("\"" + str(query.session) + "\" has released the iterate_wopfile_yml_dic_and_insert_rules_in_db lock on SQLManager.")
        return result

    def execute(self, session, statement, *args, **kwargs):
        """
        Allow to execute a statement object on the given session.

        :param session: SQLAlchemy session object.
        :param statement: SQLAlchemy statement object.
        """
        Logger.instance().debug("SQLManager.execute(" + str(session) + ", " + str(statement) + ", " +
                                    str(args) + ", " + str(kwargs) + ")")
        try:
            Logger.instance().debug(str(session) + " want the write lock on SQLManager for statement \"" + str(statement) + "\"")
            self.__lock.acquire_write()
            Logger.instance().debug(str(session) + " has taken the write lock on SQLManager.")
            return session.execute(statement, *args, **kwargs)
        finally:
            # Always release the lock
            self.__lock.release()
            Logger.instance().debug(str(session) + " has released the write lock on SQLManager.")

    def rollback(self, session):
        """
        Rollback the given session.

        The SQLManager wrap the sqlite queue for rollbacking operations on database in order to do not trigger the error due
        to the mtime_epoch_millis-out operations.

        This is done thanks to a read_write_lock: sqlmanager is a synchronized singleton with synchronized methods.

        :param session: SQLAlchemy session object
        """
        try:
            Logger.instance().debug(str(session) + " want the write lock on SQLManager.")
            self.__lock.acquire_write()
            Logger.instance().debug(str(session) + " has taken the write lock on SQLManager.")
            session.rollback()
            Logger.instance().debug(str(session) + " has been rollbacked.")
        finally:
            # Always release the lock
            self.__lock.release()
            Logger.instance().debug(str(session) + " has released the write lock on SQLManager.")

    def drop_all(self):
        """
        Use the declarative Base to drop all tables found. Should only be used for testing purposes.
        """
        try:
            self.__lock.acquire_write()
            Logger.instance().debug("Dropping all tables...")
            Base.metadata.drop_all(self.engine)
        finally:
            # Always release the lock
            self.__lock.release()

    def create_all(self):
        """
        Use the declarative Base to create all tables found.

        If you want to create a table, you should be sure that it has been imported first.
        """
        try:
            self.__lock.acquire_write()
            Logger.instance().debug("Creating all tables...")
            Base.metadata.create_all(self.engine)
            # Always release the lock
        finally:
            self.__lock.release()

    def create(self, tablename):
        """
        Use the declarative Base to create a table from its tablename.

        The tablename is the is_input of the base represented in the databse (independent of the Table model)
        :param tablename:
        """
        try:
            self.__lock.acquire_write()
            Logger.instance().debug("SQLManager.create(" + str(tablename) + "): create table " + str(tablename))
            Base.metadata.tables[tablename.split(".")[-1]].create(self.engine, checkfirst=True)
        finally:
            # Always release the lock
            self.__lock.release()

    def drop(self, tablename):
        """
        Use the declarative Base to drop a table from its tablename.

        The tablename is the is_input of the base represented in the databse (independent of the Table model)
        :param tablename:
        """
        try:
            self.__lock.acquire_write()
            # todo tabling
            Logger.instance().debug("SQLManager.drop(" + str(tablename) + "): drop table " + str(tablename))
            Base.metadata.tables[tablename.split(".")[-1]].drop(self.engine, checkfirst=True)
        finally:
            # Always release the lock
            self.__lock.release()

    def drop_table_list(self, list_str_table):
        """
        Remove a list of tables from the list of their tablenames.

        :param list_str_table: [String] the is_input of the tables.
        """
        # Get the list of Table objects from tablenames, then sort them according to their relationships / foreignkeys
        # and take the reverse to delete them in the right order (reverse of the correct order for creating them)
        # todo tabling
        list_obj_table = reversed(sort_tables([Base.metadata.tables[tablename.split(".")[-1]] for tablename in list_str_table]))
        try:
            self.__lock.acquire_write()
            for t in list_obj_table:
                Logger.instance().debug("SQLManager.drop_table_list(" + str(list_str_table) + "): drop table " + str(t.name))
                t.drop(self.engine, checkfirst=True)
        finally:
            # Always release the lock
            self.__lock.release()

    def drop_table_content_list(self, list_str_table):
        """
        Remove a list of tables from the list of their tablenames.

        :param list_str_table: [String] the is_input of the tables.
        """
        session = self.get_session()
        # Get the list of Table objects from tablenames, then sort them according to their relationships / foreignkeys
        # and take the reverse to delete them in the right order (reverse of the correct order for creating them)
        list_obj_table = reversed(
            sort_tables([Base.metadata.tables[tablename.split(".")[-1]] for tablename in list_str_table]))
        for t in list_obj_table:
            Logger.instance().debug(
                "SQLManager.drop_table_content_list(" + str(list_str_table) + "): drop table content " + str(t.name))
            self.execute(session._session(), t.delete())

    def create_trigger(self, table, ddl):
        """
        Create a Trigger after creating a given table.

        :param table:
        :type table: `Table <http://docs.sqlalchemy.org/en/latest/core/metadata.html#sqlalchemy.schema.Table>`_
        :param ddl:
        :type ddl: `DDL <http://docs.sqlalchemy.org/en/latest/core/ddl.html#sqlalchemy.schema.DDL>`_
        """
        try:
            self.__lock.acquire_write()
            event.listen(table, 'after_create', ddl)
        finally:
            self.__lock.release()




import os
import shutil
import unittest
from unittest import TestCase

from wopmars.SQLManager import SQLManager
from wopmars.WorkflowManager import WorkflowManager
from wopmars.utils.OptionManager import OptionManager
from wopmars.utils.PathManager import PathManager
from wopmars.utils.WopMarsException import WopMarsException


class TestWorkflowManager(TestCase):

    def setUp(self):

        OptionManager.initial_test_setup()  # Set tests arguments
        SQLManager.instance().create_all()  # Create database with tables
        self.test_path = PathManager.get_test_path()

        self.__s_path_to_example_definition_file_finishing = os.path.join(
            self.test_path, "resource/wopfile/example_def_file1.yml")
        self.__s_path_to_example_definition_file_that_end_with_error = os.path.join(
            self.test_path, "resource/wopfile/example_def_file5_never_ready.yml")

        self.__workflow_manager = WorkflowManager()

    def tearDown(self):

        SQLManager.instance().get_session().close()
        SQLManager.instance().drop_all()
        shutil.rmtree(os.path.join(self.test_path, "outdir"), ignore_errors=True)
        OptionManager._drop()
        SQLManager._drop()

    def test_run(self):

        # OptionManager.instance()["--dot"] = None
        #
        # OptionManager.instance()["--wopfile"] = self.__s_path_to_example_definition_file_finishing
        # with self.assertRaises(SystemExit):
        #     self.__workflow_manager.run()

        OptionManager.instance()["--wopfile"] = self.__s_path_to_example_definition_file_that_end_with_error
        with self.assertRaises(WopMarsException):
            self.__workflow_manager.run()

if __name__ == '__main__':
    unittest.main()
